/*
  AweMUD - char.h
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#ifndef CHAR_H
#define CHAR_H

#include <algorithm>

#include "entity.h"
#include "types.h"
#include "strings.h"
#include "color.h"
#include "gametime.h"
#include "ai.h"
#include "object.h"

#define MIN_NAME_LEN 4
#define MAX_NAME_LEN 15

/* C_FLAG types */
#define C_FLAG_NONE 0		// normal flags
#define C_FLAG_DEAD 1		// dead

/* tempory Define's - should be per server or per character configurable */
#define HEAL_TICKS 25

/* for get_comm_obj */
typedef enum get_obj_comm_type {
	GOC_ANY = 0,
	GOC_HELD = 1,
	GOC_ROOM = 2,
	GOC_WORN = 4
};

enum stat_spec {
	STAT_STRENGTH = 0,
	STAT_AGILITY,
	STAT_FORTITUDE,
	STAT_INTELLECT,
	STAT_SPIRIT,
	STAT_CHARISMA,
	NUM_STATS
};
enum position_spec {
	C_STAND = 0,
	C_SIT,
	C_LAY,
	C_KNEEL
};

/* stat */
struct char_stat {
	uint base;
	uint mod;
};

/* character control */
class
character : public entity
{
	E_TYPE(character)

	protected:
	/* AI */
	class npc_ai *ai_module;

	/* health */
	int health;
	int health_max;
	uint heal_ticks;

	/* character statistics */
	position_spec pos;
	char_stat stats[NUM_STATS];
	uint level;

	/* flags */
	flags_type char_flags;

	/* round_time in ticks */
	uint round_time;

	private:
	/* private io functions */
	char *get_in (void);

	/* special/control */
	void clear (void);

	void send (void);
	void receive (void);

	public:
	character (void);
	~character (void);

	/* character body */
	class char_body *body;

	/* save/load */
	virtual void save (xmlNodePtr ) const;
	virtual int load (const xmlNodePtr );

	virtual void save (void) const = 0;

	/* character info */
	position_spec get_pos (void) const { return pos; }
	int get_hp (void) const { return health; }
	int get_max_hp (void) const { return health_max; }

	/* check data */
	bool is_dead (void) const { return CHECK_FLAG (char_flags, C_FLAG_DEAD); }

	/* stats */
	uint get_stat (uint stat) const { if (stat < NUM_STATS) return stats[stat].mod + stats[stat].base; else return 0; }
	void set_stat (uint stat, int val) { if (stat < NUM_STATS) stats[stat].base = val; }
	void set_stat_mod (uint stat, int val) { if (stat < NUM_STATS) stats[stat].mod = val; }

	/* experience */
	uint get_level (void) const { return level; }

	/* player stuff */
	virtual uint get_age (void) const = 0;

	/* health */
	void heal (int amount);
	bool damage (int amount); /* returns true if damage killed */
	virtual void death (void) = 0;

	/* character abilities */
	bool can_move (void) const {
		return (
		 	(pos == C_STAND) &&
			(!is_dead ())
		);
	}
	bool can_see (void) const {
		return (
			true
		);
	}
	bool can_talk (void) const {
		return (
		 	(!is_dead ())
		);
	}
	bool can_act (void) const {
		return (
		 	(round_time == 0) &&
			(!is_dead ())
		);
	}

	/* action */
	uint add_rt (uint ticks) { return round_time += ticks; }
	uint get_rt (void) const { return round_time; }
	uint get_rts (void) const;
	uint show_rt (uint ticks);
	void fail_rt (const char *msg = NULL);

	/* char flags stuff */
	flags_type get_char_flags (void) const { return char_flags; }
	void set_char_flag (int new_flag) { SET_FLAG (char_flags, new_flag); }
	void clear_char_flag (int del_flag) { CLEAR_FLAG (char_flags, del_flag); }

	/* input/output */
	virtual const char *ncolor (void) const { return CNPC; }
	virtual void print (const char *, ...) {};
	virtual void rprint (const char *, ...);	/* shows text to char's room, but not char */
	void process_cmd (const char *);

	/* command processing utility funcs */
	class object *get_obj_comm (char *, int, bool silent = false); /* will mangle the name argument */

	/* go utility functions */
	bool enter_room (class room *, const class room_exit *);

	/* update the character */
	void update (void);

	/* remove from parent room */
	void remove (void);

	void do_emote (char const *);
	void do_say (char const *);

	void do_look (void);
	void do_look (character *);
	void do_look (const class object *, int);
	void do_look (class room_exit *);

	void do_move (int dir_t);

	void do_position (position_spec);

	void do_get (class object *, const char *, object_list *);
	void do_put (class object *, class object *, int);
	void do_drop (class object *);

	void do_wear (class object *);
	void do_remove (class object *);

	void do_read (class object *);
	void do_raise (class object *);

	void do_open (class room_exit *);
	void do_close (class room_exit *);
	void do_unlock (class room_exit *);
	void do_lock (class room_exit *);

	void do_go (const class room_exit *);

	/* for the edit extension */
	friend class edit_handler_character;

	/* parsing */
	virtual void parse_char (int c, char *buf) const;
};

/* character_list */
typedef entity_list<character> character_list;

/* utility funcs */
bool valid_name (const char *);

extern const char *stat_names[];
extern const char *stat_levels[];

const char *get_stat_level (uint);

#define CHARACTER(ent) E_CAST(ent,character)

#endif
