/*
  AweMUD - entity.h
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#ifndef ENTITY_H
#define ENTITY_H

#include <list>
#include <algorithm>

#include "types.h"
#include "error.h"
#include "mem.h"
#include "resource.h"
#include "strings.h"
#include "xml.h"
#include "log.h"
#include "registry.h"

/* EN types */
	#define EN_NONE 0			/* normal flags */
	#define EN_INV 1			/* can be inventoried */
	#define EN_MOVE 2			/* can move */
	#define EN_VISIBLE 4		/* is visible */

#define E_TYPE(name) \
	public: \
	static const void *get_setype (void) { return (void*)name::get_setype; } \
	virtual const void *get_etype (void) const { return name::get_setype (); }
#define E_CAST(ent,name) ((name *)e_cast (ent, name::get_setype ()))

class entity_property {
	protected:
	class arg_type {
		public:
		string_t name;
		string_t val;

		string_t get_name (void) const { return name; }
	};

	string_t name;
	registry<arg_type> args;

	public:
	int load (const xmlNodePtr);
	void save (xmlNodePtr) const;

	string_t get_value (const char *name) const {
		return_val_if_fail (name != NULL, "");
		const arg_type *at = args.find (name);
		if (at != NULL)
			return at->val;
		else
			return (char *)"";
	}
	int get_ivalue (const char *name) {
		return_val_if_fail (name != NULL, 0);
		return satol (get_value (name).c_str ());
	}
	void set_value (const char *name, const char *value) {
		return_if_fail (name != NULL);
		return_if_fail (value != NULL);
		arg_type *at = args.find (name);
		if (at != NULL)
			at->val = value;
		else {
			at = new arg_type ();
			at->name = name;
			at->val = value;
			args.add (at);
		}
	}
	void set_ivalue (const char *name, int i) {
		return_if_fail (name != NULL);
		set_value (name, stringf ("%d", i).c_str ());
	}

	string_t get_name (void) const { return name; }
};

class entity_event : public entity_property {
	protected:
	int action;

	public:
	int load (const xmlNodePtr);
	void save (xmlNodePtr) const;

	int get_action (void) const { return action; }
};

/* entity control */
class entity : public resource
{
	E_TYPE(entity)

	public:
	/* global list of all entities */
	static list <class entity *> entities;

	protected:
	string_t name;
	string_t desc;
	flags_type flags;
	class entity *par_ent;

	vector<entity_property *> properties;
	vector<entity_event *> events;

	int perform_event (entity_event *ea, entity *trigger, entity *target);

	entity *e_next;
	static entity *e_list;

	public:
	entity (void);
	~entity (void);

	/* can be public - just strings */
	vector<string_t> tags;

	/* flags stuff */
	int get_ent_flags (void) const { return flags; }
	void set_ent_flag (int new_flag) { SET_FLAG (flags, new_flag); }
	void clear_ent_flag (int del_flag) { CLEAR_FLAG (flags, del_flag); }

	/* tags */
	bool has_tag (string_t tag) { return std::find (tags.begin (), tags.end (), tag) != tags.end (); }
	void add_tag (string_t tag) { if (!has_tag (tag)) tags.push_back (tag); }
	void del_tag (string_t tag) { if (has_tag (tag)) tags.erase (std::find (tags.begin (), tags.end (), tag)); }

	/* get parent */
	class entity *get_par (void) const { return par_ent; }

	/* name stuff */
	string_t get_name (void) const { return name; }
	void set_name (const char *new_name) { name = new_name; }

	string_t get_desc (void) const { return desc; }
	void set_desc (const char *new_desc) { desc = new_desc; }

	/* save/load */
	virtual void save (xmlNodePtr) const;
	virtual int load (const xmlNodePtr);

	virtual void save (void) const = 0;

	/* check name */
	bool name_match (const char *name) const;

	/* event triggers */
	bool can_event (const char *) const;
	int do_event (const char *, entity *trigger, entity *target);

	/* properties */
	int add_property (entity_property *);
	int del_property (const char *);

	entity_property *get_property (const char *name) const;
	entity_event *get_event (const char *name) const;

	/* for parsing, pull a property based on a char */
	virtual void parse_char (int, char *buf) const;

	/* update */
	virtual void update (void) = 0;

	static void update_all (void);
};

template <class ENT>
class entity_list
{
	public:
	struct item {
		ENT *ent;
		item *prev;
		item *next;
	};

	protected:
	item *items, *last;

	public:
	entity_list (void) : items (NULL), last (NULL) {}
	~entity_list (void)
	{
		item *n;
		while (items != NULL) {
			n = items->next;
			items->ent->unref ();
			delete (items);
			items = n;
		}
	}

	item *entries (void) { return items; }
	const item *entries (void) const { return items; }
	
	ENT *add (ENT *ent)
	{
		return_val_if_fail (ent != NULL, NULL);

		item *i = new item ();
		i->ent = ent;
		
		if (last) {
			i->prev = last;
			i->next = NULL;
			last->next = i;
			last = i;
		} else {
			i->next = NULL;
			i->prev = NULL;
			items = i;
			last = i;
		}
		ent->ref ();
		return ent;
	}

	void remove (ENT *ent)
	{
		return_if_fail (ent != NULL);

		item *i;
		for (i = items; i != NULL; i = i->next) {
			if (i->ent == ent) {
				if (i->prev)
					i->prev->next = i->next;
				else
					items = i->next;
				if (i->next)
					i->next->prev = i->prev;
				else
					last = i->prev;
				ent->unref ();
				delete (i);
				break;
			}
		}
	}

	ENT *find (const char *str, uint c = 0)
	{
		return_val_if_fail (str != NULL, NULL);

		for (item *i = items; i != NULL; i = i->next)
			if (!strcasecmp (str, i->ent->get_name ()) && (c --) == 0)
				return i->ent;

		return NULL;
	}

	ENT *match (string_t str, uint c = 0)
	{
		return_val_if_fail (str, NULL);

		for (item *i = items; i != NULL; i = i->next)
			if (phrase_match (i->ent->get_name (), str) && (c --) == 0)
				return i->ent;

		return NULL;
	}
};

inline entity *e_cast (entity *ent, const void *check);
inline const entity *e_cast (const entity *ent, const void *check);

#define ENTITY(ent) E_CAST(ent,entity)

extern const char *entity_actions[];

#endif
