/*
  AweMUD - players.h
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#ifndef PLAYER_EXT_H
#define PLAYER_EXT_H

#include "xml.h"
#include "char.h"
#include "class.h"
#include "race.h"

#define P_FLAG_NONE 0
#define P_FLAG_VALID 1		// is completely valid
#define P_FLAG_GM 2		// is a GM
#define P_FLAG_ADMIN 4		// is admin
#define P_FLAG_BUILDER 8	// is a builder

#define CONNECTION_IN_BUFFER_LENGTH 256
#define CONNECTION_OUT_BUFFER_LENGTH 2048

/* CIO types */
#define CIO_FLAG_NONE 0			// normal flags
#define CIO_FLAG_USE_ANSI 1		// use ANSI control sequences
#define CIO_FLAG_NEED_UPDATE (1 << 2)	// need an update to menu options
#define CIO_FLAG_DEAF (1 << 3)		// hears nothing : print no text - all NPC's
#define CIO_FLAG_NEED_NEWLINE (1 << 4)	// does a newline need to be printed out?
#define CIO_FLAG_DO_ECHO (1 << 5)
#define CIO_FLAG_DO_EOR (1 << 6)
#define CIO_FLAG_WANT_ECHO (1 << 7)
#define CIO_FLAG_BT (1 << 8)

/* input states */
#define ISTATE_TEXT 0
#define ISTATE_IAC 1
#define ISTATE_WILL 2
#define ISTATE_WONT 3
#define ISTATE_DO 4
#define ISTATE_DONT 5
#define ISTATE_SB 6
#define ISTATE_SE 7
#define ISTATE_CR 8
#define ISTATE_ESC 9

enum exp_spec {
	EXP_GENERAL = 0,
	EXP_WARRIOR,
	EXP_ROGUE,
	EXP_CASTER,
	NUM_EXPS
};

class processor
{
	public:
	processor (void) {}
	virtual ~processor (void) {}

	virtual int init (class player *) { return 0; }
	virtual void finish (class player *) {}
	/* return true/non-zero in order to remove the processor */
	virtual int process (class player *, char *line) { return true; }
	virtual const char *prompt (class player *) { return "Prompt:"; }
};

class player : public character
{
	E_TYPE(player);

	protected:
	/* socket */
	int sock;

	uint8 istate;
	int16 width, out_col, in_cnt, out_cnt;
	char input_buf[CONNECTION_IN_BUFFER_LENGTH];
	char output_buf[CONNECTION_OUT_BUFFER_LENGTH];
	std::vector <uint8> sb_buf;
	string_t ip;
	uint16 port;
	time_t in_stamp;

	uint exp[NUM_EXPS];
	const char_race *race;

	struct class_data {
		uint level;
		const char_class *klass;
	};
	typedef vector<class_data> class_list;
	class_list classes;

	processor *proc;

	/* processing */
	void process_input (void);
	void process_sb (void);
	void process_comm (void);

	void send_iac (uint, ...);
	void send_data (uint, ...);

	int check_time (void); /* check to see if we should disconnect */

	void add_output (const char *, int = -1);

	/* player stuff */
	uint last_rt;
	game_time birthday;
	flags_type ply_flags;

	public:
	string_t last_command;
	string_t last_tell;
	string_t kill_name;
	string_t prompt;
	flags_type io_flags;

	/* global list - unfortunately, it must be public */
	player *p_next;

	player (void);
	~player (void);

	void save (xmlNodePtr) const;
	void save (void) const;

	int load (const xmlNodePtr);
	int load (const char *);

	/* get info */
	bool is_valid (void) const { return CHECK_FLAG (ply_flags, P_FLAG_VALID); }
	const char *ncolor (void) const { return CPLAYER; }

	/* privs */
	bool is_gm (void) const { return CHECK_FLAG (ply_flags, P_FLAG_GM); }
	bool is_admin (void) const { return CHECK_FLAG (ply_flags, P_FLAG_ADMIN); }
	bool is_builder (void) const { return CHECK_FLAG (ply_flags, P_FLAG_BUILDER); }
	/* grant permissions */
	void grant_admin (void) { SET_FLAG(ply_flags, P_FLAG_ADMIN); }
	void grant_gm (void) { SET_FLAG(ply_flags, P_FLAG_GM); }
	void grant_builder(void) { SET_FLAG(ply_flags, P_FLAG_BUILDER); }

	/* commands */
	virtual void process_cmd (const char *);

	/* misc */
	uint get_age (void) const;
	void death (void);
	void update (void);
	uint get_exp (uint type) const { if (type < NUM_EXPS) return exp[type]; else return 0; }
	void grant_exp (uint type, uint amount);
	uint get_class_level (const char_class *) const;
	void grant_class_levels (const char_class *, uint);

	/* info */
	void self_info (void);
	void show_inventory (void);
	void show_stats (void);

	/* IO options */
	bool toggle_echo (bool e);
	bool use_color (void) const { return CHECK_FLAG (io_flags, CIO_FLAG_USE_ANSI); }

	/* handle IO */
	void put (const char *);
	void print (const char *, ...);
	void rprint (const char *, ...);
	int connect (int);
	int disconnect (void);
	bool is_alive (void) const { return sock >= 0; }
	const int get_sock (void) const { return sock; }
	void process (void);
	void handle_output (void); /* deprecate me */
	void need_update (void) { SET_FLAG (io_flags, CIO_FLAG_NEED_UPDATE); }
	int add_processor (processor *);

	friend class login_proc;
	friend class create_proc;
};

/* handle login */
class login_proc : public processor
{
	int state;
	public:

	int init (player *) { state = 0; return 0; }
	int process (player *, char *);
	const char *prompt (player *);
};

/* handle changing of password */
class password_proc : public processor
{
	string_t new_pass;
	bool valid;

	public:

	int init (player *ply) { valid = false; ply->toggle_echo (false); return 0; }
	void finish (player *ply) { ply->toggle_echo (true); }
	int process (player *, char *);
	const char *prompt (player *);
};

/* handle creation of character */
class create_proc : public processor
{
	string_t pass;
	uint tokens;
	uint state;

	public:

	int init (player *ply);
	void finish (player *ply);
	int process (player *, char *);
	const char *prompt (player *);

	void start_state (player *ply);
	void process_state (player *ply, char *);
	void end_state (player *ply);
};

namespace players
{
	bool exists (const char *name);
	bool check_pass (const char *name, const char *pass);
	int set_pass (const char *name, const char *pass);
	player *load (const char *name);
	int save (const char *name, const player *ch);
	int remove (const char *name);
	player *find (const char *name);
	void list_all (player *ch);
	void process_all (void);
	void flush_all (void);
	uint count (void);
}

#define PLAYER(ent) E_CAST((ent),player)

extern const char *exp_names[];

#endif
