/*
  AweMUD - room.h
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#ifndef MAP_H
#define MAP_H

#include "mem.h"
#include "types.h"
#include "object.h"
#include "area.h"
#include "char.h"
#include "strings.h"
#include "xml.h"
#include "regname.h"

/* flags for exits */
#define EXIT_HIDDEN 1
#define EXIT_CLOSED 2
#define EXIT_NOSYNC 4
#define EXIT_LOCKED 8

/* the main directions */
typedef enum { NORTH, EAST, SOUTH, WEST, SOUTHEAST, SOUTHWEST, NORTHEAST, NORTHWEST, DIR_COUNT } dir_t;

/* direction and exit names */
extern name_registry exit_types;
extern const_name_id exit_north_id;
extern const_name_id exit_east_id;
extern const_name_id exit_south_id;
extern const_name_id exit_west_id;
extern const_name_id exit_southeast_id;
extern const_name_id exit_southwest_id;
extern const_name_id exit_northeast_id;
extern const_name_id exit_northwest_id;
extern const_name_id exit_up_id;
extern const_name_id exit_down_id;
extern const_name_id exit_door_id;
extern const_name_id exit_portal_id;
extern const_name_id exit_climb_id;
extern const_name_id exit_stairs_id;
extern const char *dir_names[];

enum {
	ROOM_EXIT_GO = 0,
	ROOM_EXIT_ENTER,
	ROOM_EXIT_CLIMB,
	ROOM_EXIT_CRAWL
};
extern const char *room_go_acts[];
extern const char *room_leave_acts[];

void init_room_exits (void);

class room_exit
{
	protected:
	string_t name;
	string_t desc;
	string_t target;
	string_t other_exit;
	string_t leave;
	string_t enter;
	string_t go;
	name_id type;
	flags_type flags;

	room_exit *next;

	public:
	room_exit (void) : type (0), flags (0), next (NULL) {}
	~room_exit (void) {}

	string_t get_name (void) const { return name; }
	void set_name (const char *n) { name = n; }

	string_t get_desc (void) const { return desc; }
	void set_desc (const char *n) { desc = n; }

	string_t get_target (void) const { return target; }
	string_t get_other_exit (void) const { return other_exit; }
	void set_target (const char *t, const char *e) { target = t; other_exit = e; }
	
	room *get_room (void) const;
	room_exit *get_exit (void) const;

	string_t get_enter (void) const { return enter; }
	void set_enter (const char *t) { enter = t; }

	string_t get_leave (void) const { return leave; }
	void set_leave (const char *t) { leave = t; }

	string_t get_go (void) const { return go; }
	void set_go (const char *t) { go = t; }

	name_id get_type (void) const { return type; }
	void set_type (name_id t) { type = t; }

	bool is_valid (void) const;
	bool is_hidden (void) const { return flags & EXIT_HIDDEN; }
	bool is_closed (void) const { return flags & EXIT_CLOSED; }
	bool is_synced (void) const { return !(flags & EXIT_NOSYNC); }
	bool is_locked (void) const { return flags & EXIT_LOCKED; }
	bool is_door (void) const { return type == exit_door_id; }

	bool handles_type (int go_type) const;
	int preferred_type (void) const;

	void lock (void);
	void unlock (void);
	void close (void);
	void open (void);

	void set_hidden (bool v) { if (!v) { flags &= ~EXIT_HIDDEN; } else { flags |= EXIT_HIDDEN; }}
	void set_closed (bool v) { if (!v) { flags &= ~EXIT_CLOSED; } else { flags |= EXIT_CLOSED; }}
	void set_synced (bool v) { if (v) { flags &= ~EXIT_NOSYNC; } else { flags |= EXIT_NOSYNC; }}
	void set_locked (bool v) { if (!v) { flags &= ~EXIT_LOCKED; } else { flags |= EXIT_LOCKED; }}

	void save (xmlNodePtr) const;
	void load (const xmlNodePtr);
	
	friend class room;
};

class room : public area
{
	E_TYPE(room)
	
	protected:
	string_t title;
	string_t short_desc;
	class zone *parent;
	bool outdoors;
	room_exit *exits;

	room *next; /* for list of rooms in world */

	public:
	flags_type flags;
	object_list objects;
	character_list chars;

	room (void);
	~room (void);

	uchar get_type (void) const { return 1; }

	bool is_outdoors (void) const { return outdoors; }
	void set_outdoors (bool v) { outdoors = v; }

	room_exit *get_exit (const char *);
	room_exit *find_exit (const char *);
	room_exit *find_exit_type (name_id); /* find based on type */

	string_t get_title (void) const { return title; }
	void set_title (const char *new_title) { title = new_title; }

	string_t get_short_desc (void) const { return short_desc; }
	void set_short_desc (const char *new_short_desc) { short_desc = new_short_desc; }

	class zone *get_parent (void) const { return parent; }
	void set_parent (class zone *new_parent) { parent = new_parent; }

	int load (const xmlNodePtr);
	void save (xmlNodePtr) const;

	int load (const char *);
	void save (void) const;

	void announce (string_t, flags_type);
	void update (void);

	void show (class character *, bool shortd = false);

	void print (const char *);
	void say (char const *, class character const *);

	class character *find_char (string_t);
	object *find_obj (string_t);

	friend class world;
};

#define ROOM(ent) E_CAST(ent,room)

#endif
