/*
  AweMUD - template.h
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#ifndef TEMPLATE_H
#define TEMPLATE_H

#include <time.h>

#include "mem.h"
#include "strings.h"
#include "error.h"
#include "xml.h"

class template_t
{
	protected:
	string_t tag;
	xmlDocPtr data;
	time_t access_time;
	unsigned int access_count;

	public:
	template_t (const char *t, xmlDocPtr d) { tag = t; data = d; access_time = time (NULL); access_count = 1; }
	template_t (void) { data = NULL; access_time = 0; access_count = 0; }
	~template_t (void)
	{
		if (data)
			xml_free (data);
	}

	int load (const xmlDocPtr info)
	{
		return_val_if_fail (info != NULL, 1);
		data = info;
		tag = xml_prop (xml_root (info), "name");
		access_time = time (NULL);
		access_count = 1;
		return 0;
	}

	unsigned long access (void) const { return (time (NULL) - access_time) / access_count; }

	const char *get_tag (void) const { return tag.c_str (); }
	const xmlDocPtr get_data (void) { access_count ++; access_time = time (NULL); return data; }
};

#define TEMPLATE_LIST_SIZE 100

class template_list_t
{
	template_t *templates[TEMPLATE_LIST_SIZE];
	string_t path;

	public:
	template_list_t (void) { for (int i = 0; i < TEMPLATE_LIST_SIZE; i ++) templates[i] = NULL; }
	~template_list_t (void) { for (int i = 0; i < TEMPLATE_LIST_SIZE; i ++) if (templates[i]) delete (templates[i]); }

	void set_path (const char *p) { path = p; }

	int free_loc (void)
	{
		int l = -1;
		unsigned long score = 0;
		unsigned long temp_score;

		for (int i = 0; i < TEMPLATE_LIST_SIZE; i ++)
		{
			if (!templates[i])
				return i;
			else
			{
				temp_score = templates[i]->access ();
				if (temp_score >= score)
				{
					score = temp_score;
					l = i;
				}
			}
		}

		/* Error... */
		if (l < 0)
		{
			warning ("Couldn't free slot in template list");
			return -1;
		}

		delete (templates[l]);
		templates[l] = NULL;

		return l;
	}

	xmlNodePtr
	get_template (const char *tag)
	{
		for (int i = 0; i < TEMPLATE_LIST_SIZE; i ++)
			if (templates[i] && str_case_equal (tag, templates[i]->get_tag ()))
				return xml_root (templates[i]->get_data ());

		string_t o_path = stringf("%s/%s.base", path.c_str (), tag);
		xmlDocPtr doc = xml_open (o_path.c_str ());
		if (!doc)
			return NULL;

		int i = free_loc ();
		if (i < 0)
			return NULL;

		template_t *temp = new template_t (tag, doc);
		if (!temp)
			return NULL;

		templates[i] = temp;
		return xml_root (temp->get_data ());
	}
};

#endif
