/*
  AweMUD - body.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <stdlib.h>

// STL stuffs
#include <algorithm>

#include "body.h"
#include "object.h"
#include "error.h"
#include "strings.h"
#include "char.h"
#include "server.h"

const char *body_loc_names[] = {
	"none",
	"head",
	"torso",
	"arm",
	"leg",
	"hand",
	"foot",
	"neck",
	"body",
	"back",
	"waist",
	NULL
};
const char *body_status_names[] = {
	"severed",
	"para",
	"scarred",
	"invisible",
	NULL
};
const char *body_pos_names[] = {
	"left",
	"right",
	"upper",
	"lower",
	"center",
	NULL
};

char_body::char_body (void)
{
	right_held = NULL;
	left_held = NULL;
	body_worn = NULL;
	waist_worn = NULL;
	back_worn = NULL;
	gen = GENDER_NONE;
}

char_body::~char_body (void)
{
	if (right_held)
		right_held->unref ();
	if (left_held)
		left_held->unref ();
	if (body_worn)
		body_worn->unref ();
	if (waist_worn)
		waist_worn->unref ();
	if (back_worn)
		back_worn->unref ();
}

void
char_body::do_load (const char *name, const xmlNodePtr config, object **obj)
{
	return_if_fail (name != NULL);
	return_if_fail (obj != NULL);

	xmlNodePtr child;
	
	if (*obj)
	{
		(*obj)->unref ();
		(*obj) = NULL;
	}

	for (child = xml_find_node (config, "part"); child != NULL; child = xml_next_node (child, "part")) {
		if (xml_prop (child, "id") && !strcmp (xml_prop (child, "id"), name)) {
			if (xml_find_node (child, "object")) {
				(*obj) = new object ();
				if ((*obj)->load (xml_find_node (child, "object"))) {
					(*obj)->unref ();
					(*obj) = NULL;
				} else
					(*obj)->ref ();
				break;
			} else if (xml_find_node (child, "o_temp")) {
				(*obj) = server.get_obj_template_copy (xml_prop (xml_find_node (child, "o_temp"), "name"));
				if (*obj)
					(*obj)->ref ();
				break;
			} else if (xml_find_node (child, "o_id")) {
				resource_id id;
				id.set_handle (xml_text (xml_find_node (child, "o_id")).c_str ());
				if (id) {
					(*obj) = new object ();
					if (*obj) {
						if ((*obj)->load (id)) {
							(*obj)->unref ();
							(*obj) = NULL;
						} else
							(*obj)->ref ();
					}
				}
			}
		}
	}
}

int
char_body::load (const xmlNodePtr config)
{
	return_val_if_fail (config != NULL, 1);

	gen = xml_prop (config, "gender", GENDER_NONE);

	do_load ("right_held", config, &right_held);
	do_load ("left_held", config, &left_held);
	do_load ("body_worn", config, &body_worn);
	do_load ("waist_worn", config, &waist_worn);
	do_load ("back_worn", config, &back_worn);

	return 0;
}

void
char_body::do_save (const char *name, xmlNodePtr config, object *obj) const
{
	return_if_fail (name != NULL);
	return_if_fail (config != NULL);

	if (obj) {
		xmlNodePtr part = xml_add_node (config, "part");
		xml_set_prop (part, "id", name);
		xml_add_text (part, "o_id", obj->get_id ().get_handle ());
		obj->save ();
	}
}

void
char_body::save (xmlNodePtr config) const
{
	return_if_fail (config != NULL);

	do_save ("right_held", config, right_held);
	do_save ("left_held", config, left_held);
	do_save ("body_worn", config, body_worn);
	do_save ("back_worn", config, back_worn);
	do_save ("waist_worn", config, waist_worn);

	xml_set_prop (config, "gender", gen);
}

bool
char_body::is_held (object *obj)
{
	return_val_if_fail (obj != NULL, false);

	if (right_held == obj)
		return true;
	if (left_held == obj)
		return true;
	return false;
}

bool
char_body::is_worn (object *obj)
{
	return_val_if_fail (obj != NULL, false);

	if (body_worn == obj)
		return true;
	if (waist_worn == obj)
		return true;
	if (back_worn == obj)
		return true;

	return false;
}

bool
char_body::is_equipped (object *obj)
{
	return_val_if_fail (obj != NULL, false);
	return is_held (obj) || is_worn (obj);
}

int
char_body::hold (object *obj)
{
	return_val_if_fail (obj != NULL, -1);

	if (right_held == NULL)
	{
		right_held = obj;
		obj->ref ();
		return 0;
	}
	if (left_held == NULL)
	{
		left_held = obj;
		obj->ref ();
		return 1;
	}
	return -1;
}

int
char_body::wear (object *obj) {
	return_val_if_fail (obj != NULL, -1);

	if (body_worn == NULL && obj->get_equip_loc () == BODYLOC_TORSO) {
		body_worn = obj;
		obj->ref ();
		return 2;
	}
	if (back_worn == NULL && obj->get_equip_loc () == BODYLOC_BACK) {
		back_worn = obj;
		obj->ref ();
		return 3;
	}
	if (waist_worn == NULL && obj->get_equip_loc () == BODYLOC_WAIST) {
		waist_worn = obj;
		obj->ref ();
		return 4;
	}

	return -1;
}

int
char_body::equip (object *obj) {
	return_val_if_fail (obj != NULL, -1);

	int ret = wear (obj);
	if (ret >= 0)
		return ret;

	return hold (obj);
}

void
char_body::let_go (object *obj) {
	return_if_fail (obj != NULL);

	if (right_held == obj) {
		right_held = NULL;
		obj->unref ();
	}
	if (left_held == obj) {
		left_held = NULL;
		obj->unref ();
	}
}

void
char_body::remove (object *obj) {
	return_if_fail (obj != NULL);

	if (body_worn == obj) {
		body_worn = NULL;
		obj->unref ();
	}
	if (back_worn == obj) {
		back_worn = NULL;
		obj->unref ();
	}
	if (waist_worn == obj) {
		waist_worn = NULL;
		obj->unref ();
	}
}

void
char_body::unequip (object *obj)
{
	return_if_fail (obj != NULL);
	let_go (obj);
	remove (obj);
}

int
char_body::hands_free (void)
{
	int hands = 0;
	if (right_held == NULL)
		hands ++;
	if (left_held == NULL)
		hands ++;
	return hands;
}

object *
char_body::get_held (unsigned int i) {
	if (right_held) {
		if (!i)
			return right_held;
		-- i;
	}

	if (left_held) {
		if (!i)
			return left_held;
		-- i;
	}

	return NULL;
}

object *
char_body::get_worn (unsigned int i)
{
	if (body_worn) {
		if (!i)
			return body_worn;
		-- i;
	}
	if (back_worn) {
		if (!i)
			return back_worn;
		-- i;
	}
	if (waist_worn) {
		if (!i)
			return waist_worn;
		-- i;
	}

	return NULL;
}

object *
char_body::get_equip (unsigned int i)
{
	object *ret;
	ret = get_held (i);
	if (ret)
		return ret;
	if (right_held) i --;
	if (left_held) i --;

	return get_worn (i);
}

object *
char_body::get_worn (string_t name)
{
	if (body_worn != NULL && phrase_match (body_worn->get_name ().c_str (), name))
		return body_worn;
	if (back_worn != NULL && phrase_match (back_worn->get_name ().c_str (), name))
		return back_worn;
	if (waist_worn != NULL && phrase_match (waist_worn->get_name ().c_str (), name))
		return waist_worn;

	return NULL;
}

object *
char_body::get_held (string_t name)
{
	if (right_held != NULL && phrase_match (right_held->get_name ().c_str (), name))
		return right_held;

	if (left_held != NULL && phrase_match (left_held->get_name ().c_str (), name))
		return left_held;

	return NULL;
}

object *
char_body::get_equip (string_t name)
{
	object *obj;
	if ((obj = get_held (name)) != NULL)
		return obj;
	else
		return get_worn (name);
}

void
char_body::swap (void) {
	object *temp = right_held;
	right_held = left_held;
	left_held = temp;
}

void
char_body::show (class character *ch)
{
	if (right_held != NULL)
		ch->print ("Right Hand ... " CITEM "%s" CNORMAL "\n", right_held->get_name ().c_str ());
	if (left_held != NULL)
		ch->print ("Left Hand ...  " CITEM "%s" CNORMAL "\n", left_held->get_name ().c_str ());
	if (body_worn != NULL)
		ch->print ("Body ...       " CITEM "%s" CNORMAL "\n", body_worn->get_name ().c_str ());
	if (back_worn != NULL)
		ch->print ("Back ...       " CITEM "%s" CNORMAL "\n", back_worn->get_name ().c_str ());
	if (waist_worn != NULL)
		ch->print ("Waist ...      " CITEM "%s" CNORMAL "\n", waist_worn->get_name ().c_str ());
}

void
char_body::drop_held (room *r) {
	return_if_fail (r != NULL);

	if (right_held) {
		r->objects.add (right_held);
		right_held->unref ();
		right_held = NULL;
	}
	if (left_held) {
		r->objects.add (left_held);
		left_held->unref ();
		left_held = NULL;
	}
}

void
char_body::drop_all (room *r) {
	return_if_fail (r != NULL);

	drop_held (r);

	if (body_worn) {
		r->objects.add (body_worn);
		body_worn->unref ();
		body_worn = NULL;
	}
	if (back_worn) {
		r->objects.add (back_worn);
		back_worn->unref ();
		back_worn = NULL;
	}
	if (waist_worn) {
		r->objects.add (waist_worn);
		waist_worn->unref ();
		waist_worn = NULL;
	}
}
