/*
  AweMUD - char.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <netinet/in.h>
#include <fnmatch.h>

#include "entity.h"
#include "char.h"
#include "error.h"
#include "server.h"
#include "strings.h"
#include "body.h"
#include "room.h"
#include "world.h"
#include "rand.h"
#include "file.h"
#include "player.h"
#include "parse.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

const char *stat_names[] = {
	"Strength",
	"Agility",
	"Fortitude",
	"Intellect",
	"Spirit",
	"Charisma",
	NULL
};
const char *stat_levels[] = {
	"Wretched",
	"Horrible",
	"Bad",
	"Poor",
	"Average",
	"Fair",
	"Good",
	"Excellent",
	"Awesome",
	NULL
};

/* ------------------ STAT INFO ---------------- */

const char *
get_stat_level (uint stat) {
	if (stat <= 15)
		return stat_levels[0];
	else if (stat <= 25)
		return stat_levels[1];
	else if (stat <= 35)
		return stat_levels[2];
	else if (stat <= 45)
		return stat_levels[3];
	else if (stat <= 55)
		return stat_levels[4];
	else if (stat <= 65)
		return stat_levels[5];
	else if (stat <= 75)
		return stat_levels[6];
	else if (stat <= 85)
		return stat_levels[7];
	else
		return stat_levels[8];
}

const char *
get_stat_color (uint stat) {
	if (stat <= 35)
		return CSTAT_BAD2;
	else if (stat <= 45)
		return CSTAT_BAD1;
	else if (stat <= 55)
		return CSTAT;
	else if (stat <= 65)
		return CSTAT_GOOD1;
	else
		return CSTAT_GOOD2;
}

/* ------------------ CHARACTER ---------------- */

void
character::save (xmlNodePtr config) const
{
	return_if_fail (config != NULL);

	entity::save (config);

	xml_set_prop (config, "position", pos);
	xml_set_prop (config, "flags", flags);

	if (CHECK_FLAG (char_flags, C_FLAG_DEAD))
		xml_set_prop (config, "dead", 1);

	xml_set_prop (config, "heal", heal_ticks);

	xml_set_prop (config, "rt", round_time);

	if (ai_module)
	{
		xmlNodePtr ai_config = xml_add_node (config, "ai");
		ai_module->save (ai_config);
	}

	xmlNodePtr health_info = xml_add_node (config, "health");
	xml_set_prop (health_info, "cur", health);
	xml_set_prop (health_info, "max", health_max);

	xmlNodePtr body_config = xml_add_node (config, "body");
	body->save (body_config);

	for (vector<string_t>::const_iterator i = tags.begin (); i != tags.end (); i ++)
		if (i->c_str ())
		{
			xmlNodePtr tag_config = xml_add_node (config, "tag");
			xml_set_text (tag_config, i->c_str ());
		}

	xmlNodePtr stats_info = xml_add_node (config, "stats");
	for (int i = 0; i < NUM_STATS; ++ i) {
		xmlNodePtr stat_info = xml_add_node (stats_info, "stat");
		xml_set_prop (stat_info, "name", stat_names[i]);
		xml_set_prop (stat_info, "value", stats[i].base);
	}
}

int
character::load (xmlNodePtr cf)
{
	return_val_if_fail (cf != NULL, 1);

	entity::load (cf);

	char_flags = 0;

	xmlNodePtr info = xml_find_node (cf, "ai");
	if (info)
	{
		ai_module = new npc_ai (this);
		ai_module->load (info);
	}

	info = xml_find_node (cf, "health");
	if (info) {
		health = xml_prop (info, "cur", 10);
		health_max = xml_prop (info, "max", 10);
	}

	info = xml_find_node (cf, "body");
	if (info)
	{
		if (body)
			delete (body);
		body = new char_body ();
		if (body->load (info))
		{
			warning ("Invalid body");
			delete (body);
			body = NULL;
		}
	}
	else
		warning ("No body");
	
	flags = xml_prop (cf, "flags", 0);
	pos = (position_spec)xml_prop (cf, "position", C_STAND);

	if (xml_prop (cf, "dead", 0))
		SET_FLAG (char_flags, C_FLAG_DEAD);

	heal_ticks = xml_prop (cf, "heal", 0);

	round_time = xml_prop (cf, "rt", 0);

	for (xmlNodePtr child = xml_find_node (cf, "tag"); child != NULL; child = xml_next_node (child, "tag"))
		tags.push_back (xml_text (child));

	for (xmlNodePtr child = xml_find_node (cf, "object"); child != NULL; child = xml_next_node (child, "object")) {
		object *obj = new object ();
		obj->ref ();
		if (!obj->load (child))
			if (body->equip (obj) < 0)
				body->hold (obj);
		obj->unref ();
	}

	for (int i = 0; i < NUM_STATS; ++ i) {
		stats[i].mod = 0;
		stats[i].base = 0;
	}
	
	xmlNodePtr stats_info = xml_find_node (cf, "stats");
	if (stats_info) {
		for (xmlNodePtr child = xml_find_node (stats_info, "stat"); child != NULL; child = xml_next_node (child, "stat")) {
			int i = get_index_of (stat_names, xml_prop (child, "name"));
			if (i >= 0)
				stats[i].base = xml_prop (child, "value", 0);
		}
	}

	return 0;
}

/* character */
character::character (void) : entity ()
{
	char_flags = C_FLAG_NONE;

	pos = C_STAND;
	ai_module = NULL;
	body = NULL;

	round_time = 0;

	heal_ticks = 0;

	for (int i = 0; i < NUM_STATS; ++ i) {
		stats[i].base = 0;
		stats[i].mod = 0;
	}

	level = 0;
}

character::~character (void)
{
	if (ai_module)
		delete ai_module;

	if (body)
		delete (body);
}

void
character::clear (void)
{
	clear ();
}

/* remove */
void
character::remove (void)
{
	if (ROOM(get_par ()))
		ROOM(get_par ())->chars.remove (this);
}

/* round time */
uint
character::get_rts (void) const {
	uint rt = round_time / (1000 / MSECS_PER_TICK);
	uint rem = round_time % (1000 / MSECS_PER_TICK);
	if (rem)
		return rt + 1;
	else
		return rt;
}

uint
character::show_rt (uint ticks) {
	round_time += ticks;
	uint rt = get_rts ();
	print ("  RT: %lu second%s\n", rt, rt > 1 ? "s" : "");
	return round_time;
}

void
character::fail_rt (const char *msg) {
	uint rt = get_rts ();
	if (msg)
		print ("You must wait %lu second%s before %s\n", rt, rt > 1 ? "s" : "", msg);
	else
		print ("You must wait %lu second%s before performing any complex actions.\n", rt, rt > 1 ? "s" : "");
}

void
character::rprint (const char *text, ...)
{
	return_if_fail (text != NULL);

	if (!ROOM(get_par ())) return;

	va_list va;
	char buf[1024];

	va_start (va, text);
	vsnprintf (buf, 1024, text, va);
	va_end (va);

	ROOM(get_par ())->print (buf);
}

/* -------------------- FUNCS ---------------------- */


bool
character::enter_room (room *new_room, const room_exit *old_exit)
{
	return_val_if_fail (new_room != NULL, false);
	if (new_room == ROOM(get_par ()))
		return true;
	bool did_enter = false;

	if (old_exit) {
		if (old_exit->get_go ())
			print ("%s\n", parse::ereplace (old_exit->get_go (), this));
		else
			print ("You %s " CREXIT "%s" CNORMAL ".\n", room_go_acts[old_exit->preferred_type ()], old_exit->get_name ().c_str ());

		if (old_exit->get_leave ())
			rprint ("%s%s" CNORMAL " %s" CNORMAL "\n", ncolor (), get_name ().c_str (), old_exit->get_leave ().c_str ());
		else
			rprint ("%s%s" CNORMAL " %s " CREXIT "%s" CNORMAL ".\n", ncolor (), get_name ().c_str (), room_leave_acts[old_exit->preferred_type ()], old_exit->get_name ().c_str ());

		if (old_exit->get_other_exit ()) {
			room_exit *new_exit = new_room->find_exit (old_exit->get_other_exit ());
			if (new_exit) {
				did_enter = true;
				if (new_exit->get_enter ())
					new_room->print (stringf ("%s%s" CNORMAL " %s" CNORMAL "\n", ncolor (), get_name ().c_str (), new_exit->get_enter ().c_str ()));
				else
					new_room->print (stringf ("%s%s" CNORMAL " arrives from " CREXIT "%s" CNORMAL ".\n", ncolor (), get_name ().c_str (), new_exit->get_name ().c_str ()));
			}
		}
	}

	if (!did_enter) {
		new_room->print (stringf ("%s%s" CNORMAL " arrives.\n", ncolor (), get_name ().c_str ()));
	}

	remove ();
	par_ent = new_room;
	new_room->chars.add (this);
	return true;
}

void
character::heal (int amount)
{
	health += amount;
	if (health > health_max)
		health = health_max;
}

bool
character::damage (int amount) {
	if (amount)
		health -= amount;
	if (health <= 0 && !is_dead ()) {
		death ();
		return true;
	} else {
		return false;
	}
}

/* ------------------ MISC STUFF ------------------- */

bool
valid_name (const char *name)
{
	if (!name)
		return 0;

	int len = strlen (name);
	if (len <= 3 || len > 14)
			return false;
	for (int i = 0; i < len; i ++)
		if (!isalpha (name[i]))
			return false;

	int name_file = open ("./data/restricted_names", O_RDONLY);
	if (name_file <= 0)
	{
		warning ("Failed to open data/restricted_names");
		return false;
	}

	string_t t_name;
	while ((t_name = read_line (name_file)).length () != 0)
	{
		message ("%s vs. %s\n", t_name.c_str (), name);
		if (!fnmatch (t_name, name, 0))
		{
			close (name_file);
			return false;
		}
	}
	close (name_file);

	return true;
}

/* ---------------- UPDATE FUNCTION----------------- */

void
character::update (void)
{
	/* update action timer */
	if (round_time > 0)
		-- round_time;

	/* healing */
	if (++heal_ticks >= HEAL_TICKS) {
		heal (1);
		heal_ticks = 0;
	}

	/* update AI */
	if (ai_module)
		ai_module->update ();
}

void
character::parse_char (int c, char *buf) const {
	switch (c) {
		case 'n':
			snprintf (buf, PARSE_RESULT_SIZE, "%s%s" CNORMAL, ncolor (), name.c_str ());
			break;
		case 'e':
			if (body)
				strcpy (buf, body->get_gender () == GENDER_MALE ? "he" : (body->get_gender () == GENDER_FEMALE ? "she" : "it"));
			break;
		case 'h':
			if (body)
				strcpy (buf, body->get_gender () == GENDER_MALE ? "him" : (body->get_gender () == GENDER_FEMALE ? "her" : "it"));
			break;
		case 's':
			if (body)
				strcpy (buf, body->get_gender () == GENDER_MALE ? "his" : (body->get_gender () == GENDER_FEMALE ? "her" : "its"));
			break;
	}
}
