/*
  AweMUD - char_comm.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include "char.h"
#include "error.h"
#include "server.h"
#include "room.h"
#include "world.h"
#include "command.h"
#include "body.h"
#include "player.h"
#include "parse.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

void
character::process_cmd (const char *line)
{
	return_if_fail (line != NULL);

	/* check for talking with \", a special case */
	if (line[0] == '\"' || line[0] == '\'')
	{
		if (line[1] != '\0')
			do_say (line+ 1);
		return;
	}

	/* check for emote'ing with :, a special case */
	if (line[0] == ':' || line[0] == ';')
	{
		if (line[1] != '\0')
			do_emote (line + 1);
		return;
	}

	commands::call (this, line);
}

object *
character::get_obj_comm (char *line, int type, bool silent)
{
	char *arg;
	bool my_flag = false;
	object *obj = NULL;

	arg = commands::get_arg (&line);
	if (!arg) {
		print ("What?\n");
		return NULL;
	}

	if (!strcasecmp (arg, "my"))
		my_flag = true;
	else
		commands::fix_arg (arg, &line);

	if (type == GOC_ANY)
		type = (get_obj_comm_type) 0xFF;

	if (my_flag || (type & GOC_ROOM) == 0)
	{
		if (type & GOC_HELD)
			obj = body->get_held (line);
		if (!obj && type & GOC_WORN)
			obj = body->get_worn (line);
		if (obj)
			return obj;
		if (!silent)
		{
			if ((type & GOC_HELD) && (type & GOC_WORN) == 0)
				print ("You are not holding '%s'.\n", line);
			else if ((type & GOC_HELD) == 0 && type & GOC_WORN)
				print ("You are not wearing '%s'.\n", line);
			else if (type & GOC_HELD && type & GOC_WORN)
				print ("You don't have '%s'.\n", line);
			else
				print ("You can't use an item in your invetory for this command.\n");
		}
		return NULL;
	}
	if (type & GOC_ROOM)
	{
		obj = ROOM(get_par ())->find_obj (line);
		if (obj)
			return obj;
	}
	if (!my_flag)
	{
		if (type & GOC_HELD)
			obj = body->get_held (line);
		if (!obj && type & GOC_WORN)
			obj = body->get_worn (line);
		if (obj)
			return obj;
	}
	print ("You do not see '%s' in the room.\n", line);

	return obj;
}

void
handle_char_move (character *ch, int dir) {
	if (ch->get_pos () == C_SIT)
		ch->print ("You cannot move while sitting down.\n");
	else if (ch->get_pos () == C_LAY)
		ch->print ("You cannot move while laying down.\n");
	else if (ch->get_pos () == C_KNEEL)
		ch->print ("You cannot move while kneeling.\n");
	else if (ch->is_dead ())
		ch->print ("You cannot move since you're dead.\n");
	else if (!ch->can_move ())
		ch->print ("You cannot move.\n");
	else {
		room_exit *exit = ROOM (ch->get_par ())->find_exit_type ((name_id)dir);
		if (exit) {
			if (exit->is_closed ()) {
				ch->print ("" CREXIT "%s" CNORMAL " is closed.\n");
			} else {
				room *new_room = exit->get_room ();
				if (new_room)
				{
					ch->enter_room (new_room, exit);
					new_room->show (ch, true); /* short desc */
				}
				else
					ch->print ("That exit does not lead anywhere.\n");
			}
		} else {
			ch->print ("You do not see an exit in that direction.\n");
		}
	}
}

namespace commands {
	/* movement commands */
	void cnorth (character *ch, char *) { handle_char_move (ch, exit_north_id); }
	void ceast (character *ch, char *) { handle_char_move (ch, exit_east_id); }
	void csouth (character *ch, char *) { handle_char_move (ch, exit_south_id); }
	void cwest (character *ch, char *) { handle_char_move (ch, exit_west_id); }
	void cnw (character *ch, char *) { handle_char_move (ch, exit_northwest_id); }
	void cne (character *ch, char *) { handle_char_move (ch, exit_northeast_id); }
	void csw (character *ch, char *) { handle_char_move (ch, exit_southwest_id); }
	void cse (character *ch, char *) { handle_char_move (ch, exit_southeast_id); }

	void cgo (character *ch, char *line)
	{
		char *arg;
		arg = get_arg (&line);
		if (arg == NULL) {
			ch->print ("Go where?\n");
			return;
		}

		const room_exit *exit;
		if ((exit = ROOM (ch->get_par ())->find_exit (arg)) != NULL) {
			if (exit->handles_type (ROOM_EXIT_GO)) {
				ch->do_go (exit);
			} else {
				ch->print ("You must %s " CREXIT "%s" CNORMAL ".\n", room_go_acts[exit->preferred_type ()], exit->get_name ().c_str ());
			}
		} else
			ch->print ("You do not see '%s'.\n", arg);
	}

	void cclimb (character *ch, char *line)
	{
		char *arg;
		arg = get_arg (&line);
		if (arg == NULL) {
			ch->print ("Climb what?\n");
			return;
		}

		const room_exit *exit;
		if ((exit = ROOM (ch->get_par ())->find_exit (arg)) != NULL) {
			if (exit->handles_type (ROOM_EXIT_CLIMB)) {
				ch->do_go (exit);
			} else {
				ch->print ("You must %s " CREXIT "%s" CNORMAL ".\n", room_go_acts[exit->preferred_type ()], exit->get_name ().c_str ());
			}
		} else
			ch->print ("You do not see '%s'.\n", arg);
	}

	void center (character *ch, char *line)
	{
		char *arg;
		arg = get_arg (&line);
		if (arg == NULL) {
			ch->print ("Enter what?\n");
			return;
		}

		const room_exit *exit;
		if ((exit = ROOM (ch->get_par ())->find_exit (arg)) != NULL) {
			if (exit->handles_type (ROOM_EXIT_ENTER)) {
				ch->do_go (exit);
			} else {
				ch->print ("You must %s " CREXIT "%s" CNORMAL ".\n", room_go_acts[exit->preferred_type ()], exit->get_name ().c_str ());
			}
		} else
			ch->print ("You do not see '%s'.\n", arg);
	}

	/* basics */
	void clook (character *ch, char *line) {
		char *arg;
		arg = get_arg (&line);
		if (arg == NULL) {
			ch->do_look ();
			return;
		}

		int contain = get_index_of (object_contain_types, arg, OBJ_CONT_NONE);
		if (contain != OBJ_CONT_NONE)
		{
			object *obj = ch->get_obj_comm (line, GOC_ANY);
			if (obj)
				ch->do_look (obj, contain);
			return;
		}

		if (!strcasecmp ("at", arg)) {
			if (!is_arg (line)) {
				ch->print ("Look at what?\n");
				return;
			}
		} else {
			fix_arg (arg, &line);
		}

		character *lch = ROOM (ch->get_par ())->find_char (line);
		if (lch)
		{
			if (lch == ch)
				ch->print ("Try info instead.\n");
			else
				ch->do_look (lch);
			return;
		}

		room_exit *e = ROOM (ch->get_par ())->find_exit (line);
		if (e)
		{
			ch->do_look (e);
			return;
		}

		object *obj = ch->get_obj_comm (line, GOC_ANY);
		if (obj)
			ch->do_look (obj, OBJ_CONT_NONE);
	}

	void csay (character *ch, char *line)
	{
		if (!is_arg (line))
			ch->print ("Say what?\n");
		else
			ch->do_say (line);
	}

	void cemote (character *ch, char *line)
	{
		if (!is_arg (line))
			ch->print ("Do what?\n");
		else
			ch->do_emote (line);
	}

	/* position */

	void cstand (character *ch, char *) { ch->do_position (C_STAND); }
	void csit (character *ch, char *) { ch->do_position (C_SIT); }
	void clay (character *ch, char *) { ch->do_position (C_LAY); }
	void ckneel (character *ch, char *) { ch->do_position (C_KNEEL); }

	/* interact */

	void copen (character *ch, char *line) {
		if (!is_arg (line)) {
			ch->print ("Open what?\n");
			return;
		}

		room_exit *exit = ROOM (ch->get_par ())->find_exit (line);
		if (exit) {
			if (exit->is_door ())
				ch->do_open (exit);
			else
				ch->print ("" CREXIT "%s" CNORMAL " is not a door.\n", exit->get_name ().c_str ());
		} else 
			ch->print ("You do not see '%s'.\n", line);
	}

	void cclose (character *ch, char *line) {
		if (!is_arg (line)) {
			ch->print ("Close what?\n");
			return;
		}

		room_exit *exit = ROOM (ch->get_par ())->find_exit (line);
		if (exit) {
			if (exit->is_door ())
				ch->do_close (exit);
			else
				ch->print ("" CREXIT "%s" CNORMAL " is not a door.\n", exit->get_name ().c_str ());
		} else 
			ch->print ("You do not see '%s'.\n", line);
	}

	void clock (character *ch, char *line) {
		if (!is_arg (line)) {
			ch->print ("Lock what?\n");
			return;
		}

		room_exit *exit = ROOM (ch->get_par ())->find_exit (line);
		if (exit) {
			if (exit->is_door ())
				ch->do_lock (exit);
			else
				ch->print ("" CREXIT "%s" CNORMAL " is not a door.\n", exit->get_name ().c_str ());
		} else 
			ch->print ("You do not see the door '%s'.\n", line);
	}

	void cunlock (character *ch, char *line) {
		if (!is_arg (line)) {
			ch->print ("Unlock what?\n");
			return;
		}

		room_exit *exit = ROOM (ch->get_par ())->find_exit (line);
		if (exit) {
			if (exit->is_door ())
				ch->do_unlock (exit);
			else
				ch->print ("" CREXIT "%s" CNORMAL " is not a door.\n", exit->get_name ().c_str ());
		} else 
			ch->print ("You do not see the door '%s'.\n", line);
	}

	void cget (character *ch, char *line) {
		const char *from[] = { "fr", "from", NULL };
		char *what, *how;
		what = get_until (&line, from);
		if (what) {
			get_arg (&line); /* clear from */
			how = get_arg (&line);
			int type = get_index_of (object_contain_types, how);
			if (type < 0) {
				type = OBJ_CONT_IN;
				fix_arg (how, &line);
			}
			if (!is_arg (line)) {
				ch->print ("Get '%s' from %s where?\n", what, object_contain_types[type]);
				return;
			}
					
			object *cobj = ch->get_obj_comm (line, GOC_ANY);
			if (!cobj)
				return;

			if (!cobj->get_container (type)) {
				ch->print ("You can't do that with '%s'.\n", cobj->get_name ().c_str ());
				return;
			}

			object *obj = cobj->get_container (type)->match (what);
			if (obj)
				ch->do_get (obj, stringf ("" CITEM "%s" CNORMAL "", cobj->get_name ().c_str ()), cobj->get_container (type));
			else
				ch->print ("Can't find '%s' %s " CITEM "%s" CNORMAL ".\n", what, object_contain_types[type], cobj->get_name ().c_str ());
		} else {
			object *obj = ch->get_obj_comm (line, GOC_ROOM);
			if (obj)
				ch->do_get (obj, NULL, NULL);
		}
	}
	
	void cput (character *ch, char *line) {
		char *what, *how;
		what = get_until (&line, object_contain_types);
		if (!what) {
			ch->print ("Put what?\n");
			return;
		}

		how = get_arg (&line);
		if (!how || !is_arg (line)) {
			ch->print ("Put '%s' where?\n", what);
			return;
		}

		int type = get_index_of (object_contain_types, how);

		object *obj = ch->get_obj_comm (what, GOC_HELD);
		if (!obj) 
			return;

		object *cobj = ch->get_obj_comm (line, GOC_ANY);
		if (!cobj)
			return;

		ch->do_put (obj, cobj, type);
	}

	void cdrop (character *ch, char *line)
	{
		if (!is_arg (line)) {
			ch->print ("Drop what?\n");
			return;
		}

		object *obj = ch->get_obj_comm (line, GOC_HELD);
		if (obj)
			ch->do_drop (obj);
	}

	void cwear (character *ch, char *line)
	{
		if (!is_arg (line)) {
			ch->print ("Wear what?\n");
			return;
		}

		object *obj = ch->get_obj_comm (line, GOC_HELD);
		if (obj)
			ch->do_wear (obj);
	}

	void cremove (character *ch, char *line) 
	{
		if (!is_arg (line)) {
			ch->print ("Remove what?\n");
			return;
		}

		object *obj = ch->get_obj_comm (line, GOC_WORN);
		if (obj)
			ch->do_remove (obj);
	}
	
	void cread (character *ch, char *line)
	{
		if (!is_arg (line)) {
			ch->print ("Read what?\n");
			return;
		}

		object *obj = ch->get_obj_comm (line, GOC_ANY);
		if (obj)
			ch->do_read (obj);
	}

	void craise (character *ch, char *line)
	{
		if (!is_arg (line)) {
			ch->print ("Raise what?\n");
			return;
		}

		object *obj = ch->get_obj_comm (line, GOC_HELD);
		if (obj)
			ch->do_raise (obj);
	}

	void cswap (character *ch, char *)
	{
		if (ch->body) {
			ch->body->swap ();
			ch->print ("You swap the contents of your hands.\n");
		} else
			ch->print ("You can't do that.\n");
	}

	/* information */

	void ctime (character *ch, char *)
	{
		ch->print ("It is currently %s on %s.  ", server.get_world ()->world_time.time_str ().c_str (), server.get_world ()->world_time.date_str ().c_str ());
		if (server.get_world ()->world_time.is_day ())
			ch->print ("It is daytime.  The Sun will set in %d hours.\n", SUN_DOWN_HOUR - server.get_world ()->world_time.hour);
		else
			ch->print ("It is nighttime.  The Sun will rise in %d hours.\n", (server.get_world ()->world_time.hour < SUN_UP_HOUR) ? SUN_UP_HOUR - server.get_world ()->world_time.hour : SUN_UP_HOUR + 24 - server.get_world ()->world_time.hour);
	}

	void cwho (player *ch, char *)
	{
		players::list_all (ch);
	}

	void cserver (character *ch, char *)
	{
		ch->print (parse::ereplace ("AweMUD V%v\nBuild: %b\nUptime: %p\n", server.get_world ()));
	}
}
