/*
  AweMUD - command.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#include <ctype.h>

#include "error.h"
#include "command.h"
#include "char.h"
#include "player.h"
#include "emotes.h"

#define DEF_COMM(name) extern void c ## name (character *, char *)
#define DEF_PCOMM(name) extern void c ## name (player *, char *)
#define DEF_GM_COMM(name) extern void gm_c ## name (player *, char *)
#define DEF_ADMIN_COMM(name) extern void admin_c ## name (player *, char *)
#define DEF_BUILDER_COMM(name) extern void builder_c ## name (player *, char *)

#define COMM(name) { #name, COMM_NA, (comm_func) c ## name }
#define PCOMM(name) { #name, COMM_PLAYER, (comm_func) c ## name }
#define GM_COMM(name) { #name, COMM_GM, (comm_func) gm_c ## name }
#define ADMIN_COMM(name) { #name, COMM_ADMIN, (comm_func) admin_c ## name }
#define BUILDER_COMM(name) { #name, COMM_BUILDER, (comm_func) builder_c ## name }

namespace commands {
	/* movement commands */
	DEF_COMM(north);
	DEF_COMM(south);
	DEF_COMM(east);
	DEF_COMM(west);

	DEF_COMM(ne);
	DEF_COMM(se);
	DEF_COMM(sw);
	DEF_COMM(nw);

	DEF_COMM(go);
	DEF_COMM(climb);
	DEF_COMM(enter);

	/* basics */

	DEF_COMM(look);
	DEF_COMM(say);
	DEF_COMM(emote);

	/* position */
	
	DEF_COMM(stand);
	DEF_COMM(sit);
	DEF_COMM(lay);
	DEF_COMM(kneel);

	/* interact */
	
	DEF_COMM(attack);
	DEF_COMM(kill);
	DEF_COMM(open);
	DEF_COMM(close);
	DEF_COMM(lock);
	DEF_COMM(unlock);
	DEF_COMM(get);
	DEF_COMM(put);
	DEF_COMM(drop);
	DEF_COMM(remove);
	DEF_COMM(wear);
	DEF_COMM(raise);
	DEF_COMM(read);
	DEF_COMM(swap);

	/* information */

	DEF_PCOMM(tell);
	DEF_PCOMM(reply);
	DEF_PCOMM(inventory);
	DEF_COMM(time);
	DEF_PCOMM(who);
	DEF_COMM(server);
	DEF_PCOMM(information);
	DEF_PCOMM(stats);
	DEF_PCOMM(password);
	DEF_PCOMM(help);
	DEF_PCOMM(prompt);

	DEF_BUILDER_COMM(load);
	DEF_BUILDER_COMM(unload);
	DEF_BUILDER_COMM(create);

	/* GM commands */

	DEF_GM_COMM(toroom);
	DEF_GM_COMM(announce);
	DEF_GM_COMM(boot);

	/* admin commands */

	DEF_ADMIN_COMM(shutdown);
	DEF_ADMIN_COMM(grant);
		
	const comm_def commands[] = {
		/* movement commands */
		COMM(north),
		COMM(south),
		COMM(east),
		COMM(west),

		COMM(ne),
		COMM(se),
		COMM(sw),
		COMM(nw),

		COMM(go),
		COMM(climb),
		COMM(enter),

		/* basics */

		COMM(look),
		COMM(say),
		COMM(emote),

		/* position */
		
		COMM(stand),
		COMM(sit),
		COMM(lay),
		COMM(kneel),

		/* interact */

		COMM(attack),
		COMM(kill),
		COMM(open),
		COMM(close),
		COMM(lock),
		COMM(unlock),
		COMM(get),
		COMM(put),
		COMM(drop),
		COMM(remove),
		COMM(wear),
		COMM(raise),
		COMM(read),
		COMM(swap),

		/* player commands */

		PCOMM(tell),
		PCOMM(reply),
		PCOMM(inventory),
		COMM(time),
		PCOMM(who),
		COMM(server),
		PCOMM(information),
		PCOMM(stats),
		PCOMM(password),
		PCOMM(help),
		PCOMM(prompt),

		/* builder command s*/
		
		BUILDER_COMM(load),
		BUILDER_COMM(unload),
		BUILDER_COMM(create),

		/* GM commands */

		GM_COMM(toroom),
		GM_COMM(announce),
		GM_COMM(boot),

		/* admin commands */

		ADMIN_COMM(shutdown),
		ADMIN_COMM(grant),

		/* end */
		{ NULL, 0, NULL }
	};

	char *
	get_arg (char **c) {
		char *arg;

		return_val_if_fail (c != NULL, NULL);

		while (isspace (**c))
			++ (*c);

		if (**c == '\0') {
			return NULL;
		}

		arg = *c;
		while (**c != '\0' && !isspace (**c))
			++ (*c);

		if (**c != '\0') {
			*((*c) ++) = '\0';
		}

		return arg;
	}

	char *
	get_until (char **c, const char **words, int* index) {
		char *arg, *start;
		int i;
		uint len;

		if (index != NULL)
			*index = -1;

		return_val_if_fail (c != NULL, NULL);
		
		while (isspace (**c))
			++ (*c);

		if (**c == '\0') {
			return NULL;
		}

		arg = *c;

		while (true) {
			start = *c;
			while (**c != '\0' && !isspace (**c))
				++ (*c);
			len = *c - start;

			for (i = 0; words[i] != NULL; ++ i) {
				if (len == strlen (words[i]) && !strncasecmp (words[i], start, len)) {
					if (start > arg)
						*(start - 1) = '\0';
					*c = start;
					if (index != NULL)
						*index = i;
					return arg;
				}
			}

			if (**c == '\0')
				break;
			else
				++ (*c);
		}

		*c = arg;
		return NULL;
	}

	bool
	is_arg (char *comm) {
		return_val_if_fail (comm != NULL, false);

		while (*comm != '\0') {
			if (!isspace (*comm))
				return true;
			++ comm;
		}
			
		return false;
	}

	char *
	fix_arg (char *arg, char **c) {
		int len = strlen (arg);
		if (arg + len == *c) {
			*c = arg;
			return arg;
		} else {
			*c = arg;
			arg[len] = ' ';
		}

		return arg;
	}

	int
	call (character *ch, const char *comm) {
		char buffer[1024];
		char *sp, *bp;
		const comm_def *cd;
		int len;
		bool sf;
		player *ply = PLAYER(ch);

		return_val_if_fail (comm != NULL, 3);

		while (*comm != '\0' && isspace (*comm))
			++ comm;

		if (*comm == '\0') {
			return 2;
		}

		sf = true;
		bp = buffer;
		while (*comm != '\0' && bp - buffer < 1023) {
			if (isspace (*comm)) {
				if (!sf) {
					*(bp ++) = ' ';
					sf = true;
				}

			} else {
				*(bp ++) = *comm;
				sf = false;
			}
			++ comm;
				
		}
		while (bp > buffer && isspace (*(bp - 1)))
			-- bp;
		*bp = '\0';

		for (sp = buffer; *sp != '\0' && !isspace (*sp); ++ sp)
			;

		if (sp == buffer) {
			return 2;
		}

		if (*sp != '\0') {
			*(sp ++) = '\0';
		}
		len = strlen (buffer);

		for (cd = commands; cd->name != NULL; ++ cd) {
			if (
				(ply == NULL && cd->priv == 0) ||
				(ply != NULL && (
					(cd->priv == 0) ||
					(cd->priv == COMM_PLAYER) ||
					(cd->priv == COMM_GM && ply->is_gm ()) ||
					(cd->priv == COMM_BUILDER && ply->is_builder ()) ||
					(cd->priv == COMM_ADMIN && ply->is_admin ())))) {
				if (!strncasecmp (cd->name, buffer, len)) {
					cd->func (ply ? ply : ch, sp);
					return 0;
				}
			}			
		}

		ch->print ("I do not understand.\n");
		return 1;
	}
}
