/*
  AweMUD - connection.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#define TELCMDS
#define TELOPTS
#define SLC_NAMES

#if !defined (_GNU_SOURCE)
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdarg.h>

#include "error.h"
#include "server.h"
#include "select.h"
#include "parse.h"
#include "player.h"

extern select_t sockets_select;

#define CONN_PUT_LEN 1024
#define CONN_TIMEOUT_MIN 30

/* helper function */
const char *
per_color (int cur, int max) {
	if (cur <= 0 || max <= 0) {
		return CSTAT_BAD2;
	} else if (max < cur) {
		return CSTAT_GOOD2;
	} else {
		int per = 100 * cur / max;
		if (per <= 20)
			return CSTAT_BAD2;
		else if (per <= 40)
			return CSTAT_BAD1;
		else if (per >= 80)
			return CSTAT_GOOD2;
		else if (per >= 60)
			return CSTAT_GOOD1;
		else
			return CSTAT;
	}
}

int
player::connect (int conn)
{
	if (sock >= 0) {
		return 1;
	}
	sock = conn;
	sockets_select.add (sock);
	input_buf[0] = '\0';
	in_cnt = 0;
	output_buf[0] = '\0';
	out_cnt = 0;
	istate = ISTATE_TEXT;
	width = -1;
	out_col = 0;
	io_flags = 0;

	SET_FLAG_TO(io_flags, CIO_FLAG_WANT_ECHO, true);
	SET_FLAG_TO(io_flags, CIO_FLAG_DO_ECHO, false);
	SET_FLAG_TO(io_flags, CIO_FLAG_DO_EOR, false);
	SET_FLAG_TO(io_flags, CIO_FLAG_USE_ANSI, true);

	struct sockaddr_in address;
	size_t addr_len = sizeof (struct sockaddr_in);
	getpeername (sock, (sockaddr *)&address, (uint *)&addr_len);
	ip = inet_ntoa (address.sin_addr);
	port = ntohs (address.sin_port);

	send_iac (2, WILL, TELOPT_EOR);
	send_iac (2, DO, TELOPT_NAWS);
	send_iac (2, DO, TELOPT_LFLOW);
	send_iac (2, DO, TELOPT_LINEMODE);

	time (&in_stamp);

	/* lock ourselves in memory */
	ref ();

	return 0;
}

/* disconnect */
int
player::disconnect (void)
{
	handle_output ();

	message (LOG_INFORMATION, "Dropped connection from %s:%d\n", ip.c_str (), port);

	if (is_valid ()) {
		remove ();
		save ();
	}

	sockets_select.remove (sock);
	close (sock);
	sock = -1;

	/* unlock from memory */
	unref ();

	return 0;
}

/* toggle echo */
bool
player::toggle_echo (bool v)
{
	SET_FLAG_TO(io_flags, CIO_FLAG_WANT_ECHO, v);
	if (v == false)
		CLEAR_FLAG(io_flags, CIO_FLAG_DO_ECHO);
	send_iac (2, v ? WONT : WILL, TELOPT_ECHO);
	return v;
}

/* put helper functions */
void
con_put (char *buf, char **cur, const char *text, int len) {
	int max = CONN_PUT_LEN - (*cur - buf);
	if (len > max)
		len = max;

	strncpy (*cur, text, len);
	*cur += len;
}
void
player::put (const char *text) 
{
	return_if_fail (text != NULL);
	const char *c;
	int len;

	if (CHECK_FLAG (io_flags, CIO_FLAG_DEAF))
		return;

	/* don't waste time if it's blank */
	if (text[0] == '\0')
		return;

	/* output a newline if we need one, such as after a prompt */
	if (CHECK_FLAG (io_flags, CIO_FLAG_NEED_NEWLINE))
	{
		add_output ("\r\n", 2);
		SET_FLAG(io_flags, CIO_FLAG_BT);
		out_col = 0;
	}
	CLEAR_FLAG(io_flags, CIO_FLAG_NEED_NEWLINE);

	do {
		/* search for newline, space, tab, or ANSI sequence */
		c = strpbrk (text, "\n\033\t ");

		/* get length of segment */
		if (c)
			len = c - text;
		else
			len = strlen (text);
		
		/* text we have is compared with outcol */
		if (width > 0 && len + out_col >= width) {
			add_output ("\r\n", 2);
			if (len) {
				add_output (text, len);
				out_col = len;
				CLEAR_FLAG(io_flags, CIO_FLAG_BT);
			} else {
				out_col = 0;
				SET_FLAG(io_flags, CIO_FLAG_BT);
			}
		} else {
			CLEAR_FLAG(io_flags, CIO_FLAG_BT);
			add_output (text, len);
			out_col += len;
		}

		if (c)
			switch (*c) {
				case ' ':
					if (!CHECK_FLAG (io_flags, CIO_FLAG_BT)) {
						add_output (" ", 1);
						out_col ++;
					}
					break;
				case '\n':
					if (!CHECK_FLAG (io_flags, CIO_FLAG_BT)) {
						add_output ("\r\n", 2);
						out_col = 0;
					}
					break;
				case '\t':
					if (!CHECK_FLAG (io_flags, CIO_FLAG_BT)) {
						add_output ("    ", out_col % 4);
						out_col += out_col % 4;
					}
					break;
				case '\033':
					/* cheat */
					text = c;
					for (; *c != '\0' && !isspace (*c) && !isalpha (*c); ++ c)
						;
					if (CHECK_FLAG (io_flags, CIO_FLAG_USE_ANSI))
						add_output (text, c - text + 1);
					break;
			}
		text = c + 1;
	} while (c != NULL);

	SET_FLAG (io_flags, CIO_FLAG_NEED_UPDATE);
}

void
player::print (char const *text, ...)
{
	return_if_fail (text != NULL);

	if (CHECK_FLAG (io_flags, CIO_FLAG_DEAF))
		return;

	va_list va;
	char buf[1024];

	va_start (va, text);
	vsnprintf (buf, 1024, text, va);
	put (buf);
	va_end (va);
}

void
player::rprint (const char *text, ...)
{
	return_if_fail (text != NULL);

	if (!ROOM(get_par ())) return;

	va_list va;
	char buf[1024];

	va_start (va, text);
	vsnprintf (buf, 1024, text, va);
	va_end (va);

	bool deaf = CHECK_FLAG (io_flags, CIO_FLAG_DEAF);
	SET_FLAG (io_flags, CIO_FLAG_DEAF);
	ROOM(get_par ())->print (buf);
	SET_FLAG_TO (io_flags, CIO_FLAG_DEAF, deaf);
}

void
player::process (void)
{
	if (!is_alive ())
		return;

	static unsigned char tempinput_buf[CONNECTION_IN_BUFFER_LENGTH];

	/* Cygwin check */
#ifdef MSG_NOSIGNAL
	int size = recv (sock, tempinput_buf, CONNECTION_IN_BUFFER_LENGTH -1, MSG_NOSIGNAL);
#else
	int size = recv (sock, tempinput_buf, CONNECTION_IN_BUFFER_LENGTH -1, 0);
#endif
	tempinput_buf[size] = 0;

	if (size < 0)
	{
		if (errno != EINTR && errno != EAGAIN)
			disconnect ();
		return;
	}
	else if (!size)
	{
		disconnect ();
		return;
	}

	time (&in_stamp);

	/* clean up */
	unsigned char c;
	for (int i = 0; i < size; i ++) {
		c = tempinput_buf[i];
		switch (istate) {
			case ISTATE_ESC:
				if (CHECK_FLAG(io_flags, CIO_FLAG_DO_ECHO))
					send_data (1, c);
				if (in_cnt < CONNECTION_IN_BUFFER_LENGTH - 1) {
					input_buf[in_cnt] = c;
					++ in_cnt;
					input_buf[in_cnt] = '\0';
				}
				istate = ISTATE_TEXT;
				break;
			case ISTATE_CR:
				istate = ISTATE_TEXT;
				if (c == '\n' || c == '\r' || c == '\0')
					break;
				/* fall thru */
			case ISTATE_TEXT:
				if (c == IAC)
					istate = ISTATE_IAC;
				else if (c == '\r' || c == '\n')
				{
					if (CHECK_FLAG(io_flags, CIO_FLAG_DO_ECHO))
						add_output ("\r\n", 2);
					process_input ();
					istate = ISTATE_CR;
				}
				else
				{
					/* to protect against "parse" hacks, this must be stripped */
					if (c != '[') {
						if (CHECK_FLAG(io_flags, CIO_FLAG_DO_ECHO))
							send_data (1, c);
						if (in_cnt < CONNECTION_IN_BUFFER_LENGTH - 1) {
							input_buf[in_cnt] = c;
							++ in_cnt;
							input_buf[in_cnt] = '\0';
						}
					}
				}
				break;
			case ISTATE_IAC:
				istate = ISTATE_TEXT;
				switch (c) {
					case WILL:
						istate = ISTATE_WILL;
						break;
					case WONT:
						istate = ISTATE_WONT;
						break;
					case DO:
						istate = ISTATE_DO;
						break;
					case DONT:
						istate = ISTATE_DONT;
						break;
					case IAC:
						break;
					case SB:
						istate = ISTATE_SB;
						break;
					case EC:
						break;
					case EL:
						input_buf[0] = '\0';
						in_cnt = 0;
						break;
				}
				break;
			case ISTATE_SB:
				if (c == IAC)
					istate = ISTATE_SE;
				else
					sb_buf.insert (sb_buf.end (), c);
				break;
			case ISTATE_SE:
				if (c == SE) {
					process_sb ();
					istate = ISTATE_TEXT;
				} else {
					sb_buf.insert (sb_buf.end (), IAC);
					if (c != IAC)
						/* I think this is incorrect, but oh well */
						sb_buf.insert (sb_buf.end (), c);
					istate = ISTATE_SB;
				}
				break;
			case ISTATE_WILL:
				istate = ISTATE_TEXT;
				switch (c) {
					case TELOPT_NAWS:
						/* ignore */
						break;
					case TELOPT_LINEMODE:
						send_iac (4, SB, TELOPT_LINEMODE, DONT, LM_FORWARDMASK);
						send_iac (1, SE);
						send_iac (4, SB, TELOPT_LINEMODE, LM_MODE, MODE_EDIT);
						send_iac (1, SE);
						break;
					default:
						send_iac (2, DONT, c);
						break;
				}
				break;
			case ISTATE_WONT:
				istate = ISTATE_TEXT;
				switch (c) {
					case TELOPT_NAWS:
						/* reset to default width */
						width = 0;
						break;
				}
				break;
			case ISTATE_DO:
				istate = ISTATE_TEXT;
				switch (c) {
					case TELOPT_ECHO:
						if (CHECK_FLAG(io_flags, CIO_FLAG_WANT_ECHO))
							SET_FLAG(io_flags, CIO_FLAG_DO_ECHO);
						break;
					case TELOPT_EOR:
						if (!CHECK_FLAG(io_flags, CIO_FLAG_DO_EOR)) {
							SET_FLAG(io_flags, CIO_FLAG_DO_EOR);
							send_iac (2, WILL, TELOPT_EOR);
						}
						break;
					default:
						send_iac (2, WONT, c);
						break;
				}
				break;
			case ISTATE_DONT:
				istate = ISTATE_TEXT;
				switch (c) {
					case TELOPT_ECHO:
						CLEAR_FLAG(io_flags, CIO_FLAG_DO_ECHO);
						break;
					case TELOPT_EOR:
						if (CHECK_FLAG(io_flags, CIO_FLAG_DO_EOR)) {
							CLEAR_FLAG(io_flags, CIO_FLAG_DO_EOR);
							send_iac (2, WONT, TELOPT_EOR);
						}
						break;
					default:
						send_iac (2, WONT, c);
						break;
				}
				break;
			/* NEWS negotiation */
			default:
				istate = ISTATE_TEXT;
				break;
		}
	}
}

void
player::process_input ()
{
	CLEAR_FLAG (io_flags, CIO_FLAG_NEED_NEWLINE);
	/* if it starts with !, process as a telnet layer command */
	if (in_cnt > 0 && input_buf[0] == '!') {
		process_comm ();
	} else {
		/* always handle quit */
		if (!strcasecmp (input_buf, "quit")) {
			put (parse::ereplace (server.quit_message (), server.get_world (), this));
			disconnect ();
			return;
		}

		if (proc) {
			if (proc->process (this, input_buf)) {
				proc->finish (this);
				delete (proc);
				proc = NULL;
			}
		} else {
			process_cmd (input_buf);
		}
	}
	input_buf[0] = '\0';
	in_cnt = 0;
}

/* process a telnet sub command */
void
player::process_sb (void)
{
	switch (sb_buf[0]) {
		/* resize of telnet window */
		case TELOPT_NAWS:
			width = sb_buf[2];
			break;
		/* negotiating linemode - oh joy */
		case TELOPT_LINEMODE:
			/* slc series */
			if (sb_buf[1] == LM_SLC) {
				/* init response */
				send_iac (3, SB, TELOPT_LINEMODE, LM_SLC);

				/* make response to each key */
				for (uint i = 2; i < sb_buf.size (); i += 3)
				{
					switch (sb_buf[i]) {
						case SLC_SYNCH:
						case SLC_AYT:
						case SLC_ABORT:
						case SLC_IP:
						case SLC_EOF:
						case SLC_SUSP:
						case SLC_AO:
							/* we don't support this crap */
							send_data (3, sb_buf[i], SLC_NOSUPPORT, 0);
							break;
						default:
							/* ACK it */
							send_data (3, sb_buf[i], sb_buf[i + 1] | SLC_ACK, sb_buf[i + 2]);
							break;
					}
				}

				/* end response command */
				send_iac (1, SE);
			} else if (sb_buf[1] == LM_MODE) {
				/* what to do here? */
			}
			break;
	}
	sb_buf.resize (0);
}

/* process commands to the telnet layer */
void
player::process_comm (void)
{
	strtok_t args (input_buf);
	const char *comm = args.strtok (" ");
	const char *arg = NULL;
	if (comm) {
		if (!strcasecmp (comm, "!color")) {
			arg = args.strtok (NULL);
			if (!strcasecmp (arg, "on")) {
				SET_FLAG(io_flags, CIO_FLAG_USE_ANSI);
				put ("" CADMIN "ANSI Color Enabled" CNORMAL "\n");
			} else if (!strcasecmp (arg, "off")) {
				CLEAR_FLAG(io_flags, CIO_FLAG_USE_ANSI);
				put ("ANSI Color Disabled\n");
			} else {
				put ("" CADMIN "ANSI Color ");
				put (CHECK_FLAG(io_flags, CIO_FLAG_USE_ANSI) ? "Enabled" : "Disabled");
				put ("" CNORMAL "\n");
			}
		} else if (!strcasecmp (comm, "!width")) {
			arg = args.strtok (NULL);
			int new_width = 0;
			if (arg)
				new_width = satol (arg);
			if (new_width >= 20)
				width = new_width;
			else if (new_width <= 0)
				width = 0;
			if (width > 0)
				put (stringf ("" CADMIN "Width: %d" CNORMAL "\n", width));
			else
				put ("" CADMIN "Width: N/A" CNORMAL "\n");
		} else if (!strcasecmp (comm, "!echo")) {
			arg = args.strtok (NULL);
			if (!strcasecmp (arg, "on")) {
				SET_FLAG(io_flags,CIO_FLAG_DO_ECHO);
				put ("" CADMIN "Echo Enabled" CNORMAL "\n");
			} else if (!strcasecmp (arg, "off")) {
				CLEAR_FLAG(io_flags,CIO_FLAG_DO_ECHO);
				put ("" CADMIN "Echo Disabled" CNORMAL "\n");
			} else {
				put ("" CADMIN "Echo ");
				put (CHECK_FLAG(io_flags, CIO_FLAG_DO_ECHO) ? "Enabled" : "Disabled");
				put ("" CNORMAL "\n");
			}
		} else if (!strcasecmp (comm, "!help")) {
			put ("Telnet commands:\n");
			put (" !color <on|off>  -- Enable or disable ANSI color.\n");
			put (" !width [value]   -- Set the column width of your display,\n");
			put ("                     as used by word-wrapping.  0 disables\n");
			put ("                     word wrapping entirely.\n");
			put (" !echo <on|off>   -- Enable or disable server echoing.\n");
		} else {
			put ("" CADMIN "Unknown Telnet Command" CNORMAL "\n");
		}
	}
}

/* flush out the output, write prompt */
void
player::handle_output (void)
{
	/* only if we're alive */
	if (!is_alive ())
		return;

	/* if we need an update to prompt, do so */
	if (CHECK_FLAG(io_flags, CIO_FLAG_NEED_UPDATE))
	{
		if (proc != NULL) {
			put (proc->prompt (this));
		} else if (prompt.length () > 0) {
			put (parse::prompt (prompt, this));
		} else {
			char prompt[64];
			snprintf (prompt, 64, "#%s%d" CNORMAL "/%s%d" CNORMAL "H %uR >", per_color (get_hp (), get_max_hp ()), get_hp (), per_color (get_hp (), get_max_hp ()), get_max_hp (), get_rts ());
			put (prompt);
		/* defautl prompt */
		}
		add_output (" ");

		/* GOAHEAD telnet command */
		if (CHECK_FLAG(io_flags, CIO_FLAG_DO_EOR))
			send_iac (1, EOR);
		
		CLEAR_FLAG (io_flags, CIO_FLAG_NEED_UPDATE);
		SET_FLAG (io_flags, CIO_FLAG_NEED_NEWLINE);
		out_col = 0;
	}

	if (out_cnt) {
		write (sock, output_buf, out_cnt);
		out_cnt = 0;
	}
}

/* send out a telnet command */
void
player::send_iac (uint count, ...)
{
	va_list va;
	static unsigned char buffer[16]; /* simple buffer */
	uint bc = 1; /* buffer index */
	uint byte;
	buffer[0] = IAC; /* we need to send an IAC */

	va_start (va, count);
	/* loop thru args */
	for (uint i = 0; i < count; i ++) {
		byte = va_arg (va, uint);

		/* add byte */
		buffer[bc ++] = byte;
		if (bc >= 16) {
			add_output ((char *)buffer, 16);
			bc = 0;
		}

		/* doube up on IAC character */
		if (byte == IAC)
			buffer[bc ++] = IAC;
		if (bc >= 16) {
			add_output ((char *)buffer, 16);
			bc = 0;
		}
	}

	/* write out rest of buffer */
	if (bc)
		add_output ((char *)buffer, bc);
}

/* send out telnet data */
void
player::send_data (uint count, ...)
{
	va_list va;
	static unsigned char buffer[16]; /* simple buffer */
	uint bc = 0; /* buffer index */
	uint byte;

	va_start (va, count);
	/* loop thru args */
	for (uint i = 0; i < count; i ++) {
		byte = va_arg (va, uint);

		/* add byte */
		buffer[bc ++] = byte;
		if (bc >= 16) {
			add_output ((char *)buffer, 16);
			bc = 0;
		}

		/* doube up on IAC character */
		if (byte == IAC)
			buffer[bc ++] = IAC;
		if (bc >= 16) {
			add_output ((char *)buffer, 16);
			bc = 0;
		}
	}

	/* write out rest of buffer */
	if (bc)
		add_output ((char *)buffer, bc);
}

void
player::add_output (const char *data, int len) {
	if (len < 0)
		len = strlen (data);

	/* keep pumping buffer out if it overflows */
	while (len + out_cnt >= CONNECTION_OUT_BUFFER_LENGTH) {
		int nlen = CONNECTION_OUT_BUFFER_LENGTH - out_cnt;
		len -= nlen;

		memcpy (output_buf + out_cnt, data, nlen * sizeof (char));
		write (sock, output_buf, CONNECTION_OUT_BUFFER_LENGTH * sizeof (char));
		out_cnt = 0;

		data += nlen;
	} 

	if (len > 0) {
		memcpy (output_buf + out_cnt, data, len * sizeof (char));
		out_cnt += len;
	}
}

int
player::check_time (void)
{
	time_t now;

	time (&now);

	if ((unsigned long)difftime (now, in_stamp) >= (CONN_TIMEOUT_MIN * 60)) {
		message (LOG_INFORMATION, "Connection timeout (%d minutes of no input) for %s:%d\n", CONN_TIMEOUT_MIN, ip.c_str (), port);
		disconnect ();
		return 1;
	}

	return 0;
}

int
player::add_processor (processor *p)
{
	return_val_if_fail (p != NULL, 1);

	if (proc) {
		proc->finish (this);
		delete (proc);
	}

	proc = p;
	if (p->init (this)) {
		delete (p);
		proc = NULL;
		return 1;
	} else {
		need_update ();
		return 0;
	}
}

namespace players
{
	extern player *ply_list;

	void
	process_all (void)
	{
		player *last, *conn;

		last = NULL;
		conn = ply_list;
		while (conn != NULL) {
			if (sockets_select.check (conn->get_sock ())) {
				conn->process (); 
				if (!conn->is_alive ()) {
					if (last) {
						last->p_next = conn->p_next;
						conn->unref ();
						conn = last->p_next;
					} else {
						ply_list = conn->p_next;
						conn->unref ();
						conn = ply_list;
					}
				} else {
					sockets_select.uncheck (conn->get_sock ());
					last = conn;
					conn = conn->p_next;
				}
			} else {
				last = conn;
				conn = conn->p_next;
			}
		}
	}

	void
	flush_all (void)
	{
		player *last, *conn;

		last = NULL;
		conn = ply_list;
		while (conn != NULL) {
			if (!conn->is_alive ()) {
				if (last) {
					last->p_next = conn->p_next;
					conn->unref ();
					conn = last->p_next;
				} else {
					ply_list = conn->p_next;
					conn->unref ();
					conn = ply_list;
				}
			} else {
				conn->handle_output(); 
				last = conn;
				conn = conn->p_next;
			}
		}
	}
}
