/*
  AweMUD - creation.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <ctype.h>

#include "error.h"
#include "server.h"
#include "rand.h"
#include "world.h"
#include "body.h"
#include "player.h"
#include "parse.h"
#include "command.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

namespace creation
{
	void
	show_races (player *ch)
	{
		ch->print ("Available races:\n");
		for (vector<char_race *>::const_iterator i = races.begin (); i != races.end (); i ++) {
			ch->print (" %s\n", (*i)->get_name ().c_str ());
		}
	}

	void
	show_classes (player *ch)
	{
		ch->print ("Available classes:\n");
		for (vector<char_class *>::const_iterator i = classes.begin (); i != classes.end (); i ++) {
			ch->print (" %s\n", (*i)->get_name ().c_str ());
		}
	}

	void
	show_genders (player *ch)
	{
		ch->print ("Available races:\n");
		ch->print (" F) Female\n");
		ch->print (" M) Male\n");
	}
}

int
create_proc::init (player *ply)
{
	ply->print ("\n\n ** CHARACTER CREATOR **\n\n");
	message (LOG_INFORMATION, "Beginning creation of %s\n", ply->get_name ().c_str ());

	tokens = 6;
	state = 0;

	ply->body = new char_body ();

	creation::show_races (ply);

	return 0;
}

const char *
create_proc::prompt (player *)
{
	static const char *prompts[] = {
		"Select your race:",
		"Select your gender:",
		"Select a stat to raise:",
		"Select your class:",
		"Enter a password:",
		"Verify your password:",
		NULL
	};

	return prompts[state];
}

void
create_proc::finish (player *ply)
{
	ply->health_max = ply->health = 20 + roll_dice (3, 10);

	ply->birthday = server.get_world ()->world_time;
	ply->birthday.year -= 20 + get_random (10);

	ply->show_inventory ();
	ply->self_info ();
	
	message (LOG_INFORMATION, "%s created\n", ply->get_name ().c_str ());
	/* show the menu choices */
	ply->print ("%s\n", parse::ereplace (server.logon_message (), ply, server.get_world ()));
	SET_FLAG (ply->ply_flags, P_FLAG_VALID);

	/* load the starting room */
	ply->enter_room (server.get_world ()->get_start_room (), NULL);
	ply->do_look ();
	players::set_pass (ply->get_name (), pass);
	ply->save ();
}

void
create_proc::start_state (player *ply)
{
	switch (state) {
		case 0:
			creation::show_races (ply);
			break;
		case 1:
			creation::show_genders (ply);
			break;
		case 2:
			tokens = NUM_STATS;
			ply->print ("You are given %d 'tokens' with which is increase your character "
					"attributes.  Each token increases an attribute by 5 points.  "
					"You may at any time reset your stats by typing 'roll'.  "
					"When you are finished, type 'done'.\n\n", tokens);
			for (int i = 0; i < NUM_STATS; ++ i) {
				ply->set_stat_mod (i, 0);
				ply->set_stat (i, 30 + get_random (41));
			}
			ply->show_stats ();
			break;
		case 3:
			creation::show_classes (ply);
			break;
		case 4:
		case 5:
			ply->toggle_echo (false);
			ply->put ("\n");
			break;
		default:
			ply->print ("Creation complete.\n");
			state = 1000;
			break;
	}
}

int
create_proc::process (player *ply, char *line)
{
	if (commands::is_arg (line)) {
		uint lstate = state;
		process_state (ply, line);
		if (state != lstate) {
			start_state (ply);
		}
		if (state >= 1000) {
			return true;
		}
	} else {
		ply->need_update ();
	}
	return false;
}

void
create_proc::process_state (player *ply, char *line)
{
	switch (state) {
		case 0: /* select race */
			{
				for (vector<char_race *>::const_iterator i = races.begin (); i != races.end (); i ++)
					if (!strncasecmp (line, (*i)->get_name (), strlen (line))) {
						ply->race = *i;
						break;
					}
				if (ply->race == NULL) {
					creation::show_races (ply);
				} else {
					ply->print ("Race will be %s.\n\n", ply->race->get_name ().c_str ());
					++ state;
				}
				break;
			}
		case 1: /* select gender */
			{
				if (strlen (line)) {
					if (!strncasecmp ("female", line, strlen (line))) {
						ply->body->set_gender (GENDER_FEMALE);
						ply->print ("Gender will be female.\n\n");
						++ state;
					} else if (!strncasecmp ("male", line, strlen (line))) {
						ply->body->set_gender (GENDER_MALE);
						ply->print ("Gender will be male.\n\n");
						++ state;
					} else {
						creation::show_genders (ply);
					}
				} else {
					creation::show_genders (ply);
				}
				break;
			}
		case 2: /* raise stats */
			{
				if (!strncasecmp ("roll", line, strlen (line))) {
					start_state (ply);
				} else if (!strncasecmp ("done", line, strlen (line))) {
					if (tokens > 0) {
						ply->print ("You still have tokens remaining.\n");
					} else {
						ply->print ("Stats finalized.\n\n");
						++ state;
					}
				} else {
					int stat = guess_index_of (stat_names, line);
					if (stat >= 0) {
						if (tokens > 0) {
							ply->print ("One token spent on '%s'.\n", stat_names[stat]);
							ply->set_stat (stat, ply->get_stat (stat) + 5);
							ply->show_stats ();
							-- tokens;
							ply->print ("You have %d tokens left.\n", tokens);
						} else {
							ply->print ("You spent all your tokens.\n");
						}
					} else {
						ply->print ("Invalid stat name.\n");
					}
				}
				break;
			}
		case 3: /* select class */
			{
				char_class *klass = NULL;
				for (vector<char_class *>::const_iterator i = classes.begin (); i != classes.end (); i ++)
					if (!strncasecmp (line, (*i)->get_name (), strlen (line))) {
						klass = *i;
						break;
					}
				if (klass == NULL) {
					creation::show_classes (ply);
				} else {
					ply->print ("Class will be %s.\n\n", klass->get_name ().c_str ());
					ply->grant_class_levels (klass, 1);
					++ state;
				}
				break;
			}
		case 4: /* select a password */
			pass = line;
			++ state;
			ply->need_update ();
			break;
		case 5:
			if (strcmp (pass, line)) {
				ply->print ("Passwords do not match.\n\n");
				-- state;
			} else {
				players::set_pass (ply->get_name (), line);
				ply->print ("Password set.\n\n");
				ply->toggle_echo (true);
				++ state;
			}
			break;
		default:
			{
				ply->print ("Invalid state, finishing creation.\n");
				state = 1000;
				break;
			}
	}
}
