/*
  AweMUD - entity.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <stdlib.h>
#include <stdio.h>
#include <algorithm>

#include "entity.h"
#include "error.h"
#include "file.h"
#include "strings.h"
#include "server.h"
#include "combat.h"
#include "scriptix.h"
#include "parse.h"

entity *entity::e_list = NULL;

const char *entity_actions[] = {
	"print",
	"spawn_obj",
	"spawn_npc",
	"script",
	NULL
};

int
entity_property::load (const xmlNodePtr info)
{
	name = xml_prop (info, "name");

	arg_type *at;
	for (xmlNodePtr child = xml_find_node (info, "value"); child != NULL; child = xml_next_node (child, "value")) {
		at = new arg_type ();
		if (at == NULL)
			error ("Out of memory");

		at->name = xml_prop (child, "name");
		at->val = xml_text (child);
		args.add (at);
	}
	return 0;
}

void
entity_property::save (xmlNodePtr info) const
{
	xml_set_prop (info, "name", name.c_str ());

	for (registry<arg_type>::iterator i = args.begin (); i != args.end (); i ++) {
		if ((*i)->name) {
			xmlNodePtr arg = xml_add_node (info, "value");
			xml_set_prop (arg, "name", (*i)->name.c_str ());
			xml_set_text (arg, (*i)->val);
		}
	}
}

int
entity_event::load (const xmlNodePtr info) {
	action = get_index_of (entity_actions, xml_prop (info, "action"), -1);

	return entity_property::load (info);
}

void
entity_event::save (xmlNodePtr info) const {
	if (action >= 0) {
		xml_set_prop (info, "action", get_text_of (entity_actions, action));

	} else {
		xml_set_prop (info, "action", "nil");
	}
	entity_property::save (info);
}

entity::entity (void) : resource (), name (), desc ()
{
	flags = 0;
	par_ent = NULL;
	e_next = e_list;
	e_list = this;
}

entity::~entity (void)
{
	if (e_list == this) {
		e_list = e_next;
	} else {
		entity *e;
		for (e = e_list; e != NULL && e->e_next != NULL && e->e_next != this; e = e->e_next)
			;
		if (e != NULL && e->e_next == this)
			e->e_next = e_next;
	}

	for (vector<entity_event *>::iterator i = events.begin (); i != events.end (); i ++)
		delete (*i);

	for (vector<entity_property *>::iterator i = properties.begin (); i != properties.end (); i ++)
		delete (*i);
}

bool
entity::name_match (const char *name) const
{
	return phrase_match (get_name (), name);
}

void
entity::save (xmlNodePtr config) const
{
	xml_set_prop (config, "name", name.c_str ());
	xml_set_prop (config, "id", id.get_handle ().c_str ());

	for (vector<string_t>::const_iterator i = tags.begin (); i != tags.end (); i ++)
		if (i->c_str ())
			xml_set_text (xml_add_node (config, "tag"), i->c_str ());

	if (desc)
		xml_add_text (config, "desc", desc);

	for (vector<entity_property *>::const_iterator i = properties.begin (); i != properties.end (); i ++)
		(*i)->save (xml_add_node (config, "property"));

	for (vector<entity_event *>::const_iterator i = events.begin (); i != events.end (); i ++)
		(*i)->save (xml_add_node (config, "event"));
}

int
entity::load (const xmlNodePtr config)
{
	name = xml_prop (config, "name");
	if (!name)
		name = xml_find_text (config, "name");
	if (!name)
		name = "unnamed entity";

	for (xmlNodePtr child = xml_find_node (config, "tag"); child != NULL; child = xml_next_node (child, "tag"))
		tags.push_back (xml_text (child));

	desc = xml_find_text (config, "desc");

	string_t id_str = xml_prop (config, "id");
	if (!id_str)
		id.init ();
	else
		id.set_handle (id_str);
	if (!id.valid ())
		id.init ();

	for (xmlNodePtr child = xml_find_node (config, "property"); child != NULL; child = xml_next_node (child, "property"))
	{
		entity_property *prop = new entity_property ();
		if (prop != NULL) {
			prop->load (child);
			properties.push_back (prop);
		}
	}

	for (xmlNodePtr child = xml_find_node (config, "event"); child != NULL; child = xml_next_node (child, "event"))
	{
		entity_event *ev = new entity_event ();
		if (ev != NULL) {
			ev->load (child);
			events.push_back (ev);
		}
	}

	return 0;
}

int
entity::add_property (entity_property *prop)
{
	return_val_if_fail (prop != NULL, 1);

	if (get_property (prop->get_name ()))
		return 2;
	else
		properties.push_back (prop);

	return 0;
}

int
entity::del_property (const char *name)
{
	return_val_if_fail (name != NULL, 1);

	for (vector<entity_property *>::iterator i = properties.begin (); i != properties.end (); i ++)
		if (!strcasecmp (name, (*i)->get_name ()))
		{
			properties.erase (i);
			return 0;
		}
	return 2;
}

bool
entity::can_event (const char *event) const {
	return get_event (event) != NULL;
}

int
entity::do_event (const char *event, entity *trigger, entity *target) {
	bool did_event = false;
	for (vector<entity_event *>::const_iterator i = events.begin (); i != events.end (); ++ i)
		if (event == (*i)->get_name () && (*i)->get_action () >= 0) {
			if (!perform_event ((*i), trigger, target))
				return 0;
			did_event = true;
		}
	return did_event ? 1 : -1;
}

/* return:
 * 	< 0: no action
 * 	> 0: success
 * 	= 0: failed to complete action, cancel action
 */
int
entity::perform_event (entity_event *ev, entity *trigger, entity *target) {
	switch (ev->get_action ()) {
		case 0:	{ /* print something */
			string_t text;
			if (trigger && CHARACTER (trigger)) {
				text = ev->get_value ("trigger");
				if (text)
					(CHARACTER (trigger))->print (text);
			}
			if (target && CHARACTER (target)) {
				text = ev->get_value ("target");
				if (text)
					(CHARACTER (target))->print ("target");
			}
			if (trigger || target) {
				text = ev->get_value ("room");
				if (text) {
					if (target && ROOM (target->get_par ()))
						ROOM (target->get_par ())->print (text);
					else if (ROOM (trigger->get_par ()))
						ROOM (trigger->get_par ())->print (text);
				}
			}
			return 1;
		}
		case 1: { /* spawn object */
			object *obj = server.get_obj_template_copy (ev->get_value ("id"));
			if (obj) {
				if (ROOM (this->get_par ())) {
					ROOM (this->get_par ())->objects.add (obj);
				} else if (ROOM (trigger->get_par ())) {
					ROOM (trigger->get_par ())->objects.add (obj);
				} else {
					obj->unref ();
				}
			}
			return 1;
		}
		case 2: { /* spawn npc */
			character *npc = server.get_char_template_copy (ev->get_value ("id"));
			if (npc) {
				if (ROOM (this->get_par ())) {
					npc->enter_room (ROOM (this->get_par ()), NULL);
				} else if (ROOM (trigger->get_par ())) {
					npc->enter_room (ROOM (trigger->get_par ()), NULL);
				} else {
					npc->unref ();
				}
			}
			return 1;
		}
		case 3: { /* run script */
			string_t script = ev->get_value ("name");
			if (script)
				run_script (script, "eee", this, trigger, target);
			return 1;
		}
	}

	return -1;
}

entity_property *
entity::get_property (const char *name) const
{
	return_val_if_fail (name != NULL, NULL);
	for (vector<entity_property *>::const_iterator i = properties.begin (); i != properties.end (); i ++)
		if (name ==  (*i)->get_name ())
			return (*i);
	return NULL;
}

entity_event *
entity::get_event (const char *name) const
{
	return_val_if_fail (name != NULL, NULL);
	for (vector<entity_event *>::const_iterator i = events.begin (); i != events.end (); i ++)
		if (name ==  (*i)->get_name ())
			return (*i);
	return NULL;
}

void
entity::parse_char (int c, char *buf) const {
	if (c == 'n') {
		snprintf (buf, PARSE_RESULT_SIZE, "%s", name.c_str ());
	}
}

entity *
e_cast (entity *ent, const void *check) {
	return ((ent != NULL) && (ent->get_etype () == check)) ? ent : NULL;
}

const entity *
e_cast (const entity *ent, const void *check) {
	return ((ent != NULL) && (ent->get_etype () == check)) ? ent : NULL;
}

void
entity::update_all (void) {
	entity *e, *n;

	for (e = e_list; e != NULL; e = n) {
		e->ref ();
		e->update ();
		n = e->e_next;
		e->unref ();
	}
}
