/*
  AweMUD - gametime.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#include "gametime.h"
#include "server.h"

void
game_time_def::load (const xmlNodePtr info)
{
	return_if_fail (info != NULL);

	for (xmlNodePtr m_info = xml_find_node (info, "month"); m_info != NULL; m_info = xml_next_node (m_info, "month"))
	{
		game_time_def::month_def month;
		month.name = xml_prop (m_info, "name");
		month.day_count = xml_prop (m_info, "days", 1);
		month.leap_years = xml_prop (m_info, "leapyears", 0);	
		months.push_back (month);
	}

	for (xmlNodePtr wd_info = xml_find_node (info, "weekday"); wd_info != NULL; wd_info = xml_next_node (wd_info, "weekday"))
	{
		string_t week_day = xml_text (wd_info);
		week_days.push_back (week_day);
	}
}

void
game_time_def::save (xmlNodePtr info) const
{
	return_if_fail (info != NULL);

	for (vector<game_time_def::month_def>::const_iterator i = months.begin (); i != months.end (); i ++)
	{
		xmlNodePtr m_info = xml_add_node (info, "month");
		xml_set_prop (m_info, "name", i->name.c_str ());
		xml_set_prop (m_info, "days", i->day_count);
		xml_set_prop (m_info, "leapyears", i->leap_years);
	}

	for (vector<string_t>::const_iterator i = week_days.begin (); i != week_days.end (); i ++)
		xml_add_text (info, "weekday", i->c_str ());
}

uint8
game_time_def::get_week_day (const game_time &gt) const
{
	uint64 total_days = 0;

	for (vector<game_time_def::month_def>::const_iterator i = months.begin (); i != months.end (); i ++)
	{
		uint year = gt.year - (gt.month > months.end () - i + 1 ? 0 : 1);
		total_days += i->day_count * year;
		if (i->leap_years && year % i->leap_years == 0)
			total_days ++;
	}
	total_days += gt.day;

	return total_days % week_days.size ();
}

uint16
game_time_def::days_in_month (const game_time &gt) const
{
	if (gt.month < 1 || gt.month > months.size ())
		return 0;
	if (months[gt.month - 1].leap_years && (gt.year % months[gt.month - 1].leap_years) == 0)
		return months[gt.month - 1].day_count + 1;
	else
		return months[gt.month - 1].day_count;
}

string_t
game_time::time_str (void) const
{
	return stringf ("%d:%02d %s",
			hour == 0 ? 12 : (hour <= 12 ? hour : hour - 12),
			ticks_in_hour * 60 / TICKS_PER_HOUR,
			(hour <= 12) ? "am" : "pm");
}

string_t
game_time::date_str (void) const
{
	return stringf ("%s, %d%s of %s, %d",
			server.get_world ()->time_def.week_days[server.get_world ()->time_def.get_week_day (*this)].c_str (),
			day,
			get_num_suffix (day),
			server.get_world ()->time_def.months[month - 1].name.c_str (),
			year);
}

void
game_time::update (uint ticks)
{
	ticks_in_hour += ticks;
	clip_time ();
}

void
game_time::clip_time (void)
{
	while (ticks_in_hour >= TICKS_PER_HOUR)
	{
		ticks_in_hour -= TICKS_PER_HOUR;
		hour ++;
	}
	while (hour >= 24)
	{
		hour -= 24;
		day ++;
	}
	while (day > server.get_world ()->time_def.days_in_month (*this))
	{
		day -= server.get_world ()->time_def.days_in_month (*this);
		month ++;
	}
	if (day < 1)
		day = 1;
	while (month > server.get_world ()->time_def.months.size ())
	{
		month -= server.get_world ()->time_def.months.size ();
		year ++;
	}
	if (month < 1)
		month = 1;
}

void
game_time::load (const xmlNodePtr info)
{
	return_if_fail (info != NULL);
	year = xml_prop (info, "year", 0);
	month = xml_prop (info, "month", 1);
	day = xml_prop (info, "day", 1);
	hour = xml_prop (info, "hour", 0);
	ticks_in_hour = xml_prop (info, "ticks", 0);
	clip_time ();
}

void
game_time::save (xmlNodePtr info) const
{
	return_if_fail (info != NULL);
	xml_set_prop (info, "year", year);
	xml_set_prop (info, "month", month);
	xml_set_prop (info, "day", day);
	xml_set_prop (info, "hour", hour);
	xml_set_prop (info, "ticks", ticks_in_hour);
}
