/*
  AweMUD - object.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <stdlib.h>

#include "entity.h"
#include "object.h"
#include "strings.h"
#include "error.h"
#include "char.h"
#include "room.h"
#include "server.h"
#include "body.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

const char *object_contain_types[] = {
	"none",
	"in",
	"on",
	"under",
	"behind",
	"up",
	"down",
	NULL
};

object::object (void) : entity ()
{
	parent = NULL;
}

/* _object save */
void
object::save (xmlNodePtr info) const
{
	return_if_fail (info != NULL);

	entity::save (info);
	
	if (flags)
		xml_set_prop (info, "flags", flags);
	if (equip_loc != BODYLOC_NONE)
		xml_set_prop (info, "equip", get_text_of (body_loc_names, equip_loc));

	if (CHECK_FLAG (object_flags, O_FLAG_HIDDEN))
		xml_set_prop (info, "hidden", 1);
	if (CHECK_FLAG (object_flags, O_FLAG_NO_GET))
		xml_set_prop (info, "noget", 1);

	for (vector<contain_type *>::const_iterator i = container.begin (); i != container.end (); i ++)
	{
		xmlNodePtr contain_info = xml_add_node (info, "container");
		xml_set_prop (contain_info, "type", get_text_of (object_contain_types, (*i)->type));
		for (const object_list::item *e = (*i)->list.entries (); e != NULL; e = e->next)
			e->ent->save (xml_add_node (contain_info, "object"));
	}
}


/* _object load */
int
object::load (const xmlNodePtr obj_info)
{
	return_val_if_fail (obj_info != NULL, 1);

	entity::load (obj_info);

	desc = xml_find_text (obj_info, "desc");

	object_flags = 0;
	if (xml_prop (obj_info, "hidden", 0))
		SET_FLAG (object_flags, O_FLAG_HIDDEN);
	if (xml_prop (obj_info, "noget", 0))
		SET_FLAG (object_flags, O_FLAG_NO_GET);

	equip_loc = (bodyloc_spec)get_index_of (body_loc_names, xml_prop (obj_info, "equip"), xml_prop (obj_info, "equip", BODYLOC_NONE));

	flags = xml_prop (obj_info, "flags", 0);

	for (xmlNodePtr child = xml_find_node (obj_info, "container"); child != NULL; child = xml_next_node (child, "container"))
	{
		contain_type *contain = new contain_type ();;
		contain->type = (object_contain_type)get_index_of (object_contain_types, xml_prop (child, "type"), xml_prop (child, "type", OBJ_CONT_IN));
		for (xmlNodePtr o_child = xml_find_node (child, "object"); o_child != NULL; o_child = xml_next_node (o_child, "object"))
		{
			object *obj = new object ();
			obj->load (o_child);
			contain->list.add (obj);
		}
		container.push_back (contain);
	}

	return 0;
}

void
object::save (void) const {
	xmlDocPtr obj_doc = xml_create ("object", "http://www.awemud.net/xml/" BASE_VERSION "/object.dtd");
	if (!obj_doc)
		return;

	save (xml_root (obj_doc));

	xml_save (obj_doc, stringf ("world/objects/%s", id.get_handle().c_str ()));

	xml_free (obj_doc);

	return;
}

int
object::load (const resource_id &id) {
	xmlDocPtr obj_doc = xml_open (stringf ("world/objects/%s", id.get_handle ().c_str ()));
	if (!obj_doc)
		return 1;

	int ok = load (xml_root (obj_doc));

	xml_free (obj_doc);

	return ok;
}

void
object::update (void)
{
}

void
object::set_parent (object *obj)
{
	if (obj)
		parent = obj;
	else
		parent = NULL;
}

object_list *
object::get_container (int type) const
{
	for (vector<contain_type *>::const_iterator i = container.begin (); i != container.end (); i ++)
		if ((*i)->type == type)
			return &(*i)->list;
	return NULL;
}

void
object::set_container_exist (int type, bool exist)
{
	for (vector<contain_type *>::iterator i = container.begin (); i != container.end (); i ++)
		if ((*i)->type == type)
		{
			if (!exist)
				container.erase (i);
			return;
		}
	if (exist)
	{
		contain_type *contain = new contain_type ();;
		contain->type = (object_contain_type)type;
		container.push_back (contain);
	}
}
