/*
  AweMUD - players.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#if !defined (_GNU_SOURCE)
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <unistd.h>
#include <ctype.h>

#include "char.h"
#include "player.h"
#include "server.h"
#include "strings.h"
#include "file.h"
#include "rand.h"
#include "body.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

const char *exp_names[] = {
	"General",
	"Warrior",
	"Rogue",
	"Caster",
	NULL
};

extern const char * get_stat_level (uint stat);
extern const char * get_stat_color (uint stat);

namespace players {
	player *ply_list = NULL;
	const char *get_path (const char *name, bool pass = false);
}

player::player (void) : character ()
{
	sock = -1;
	p_next = players::ply_list;
	players::ply_list = this;
	io_flags = 0;
	ply_flags = 0;
	last_rt = 0;

	race = NULL;

	for (int i = 0; i < NUM_EXPS; ++ i) {
		exp[i] = 0;
	}

	/* login */
	proc = new login_proc ();
	proc->init (this);

	/* added to list, must ref ()*/
	ref ();
}

player::~player (void)
{
}

void
player::save (xmlNodePtr config) const
{
	character::save (config);

	xmlNodePtr player_config = xml_add_node (config, "player");

	birthday.save (xml_add_node (player_config, "birthday"));

	if (CHECK_FLAG (ply_flags, P_FLAG_GM))
		xml_set_prop (player_config, "gm", 1);
	if (CHECK_FLAG (ply_flags, P_FLAG_BUILDER))
		xml_set_prop (player_config, "builder", 1);
	if (CHECK_FLAG (ply_flags, P_FLAG_ADMIN))
		xml_set_prop (player_config, "admin", 1);
	
	if (race != NULL) 
		xml_set_prop (config, "race", race->get_name ().c_str ());

	if (prompt.length () > 0)
		xml_add_text (player_config, "prompt", prompt);

	/* save player info */
	if (ROOM (get_par ())) 
		xml_set_prop (player_config, "room", ROOM (get_par ())->get_name ().c_str ());

	xmlNodePtr exp_info = xml_add_node (config, "exp");
	for (int i = 0; i < NUM_EXPS; ++ i) {
		xmlNodePtr e_info = xml_add_node (exp_info, "eset");
		xml_set_prop (e_info, "name", exp_names[i]);
		xml_set_prop (e_info, "amount", exp[i]);
	}

	xmlNodePtr classes_info = xml_add_node (config, "classes");
	for (class_list::const_iterator i = classes.begin (); i != classes.end (); ++ i) {
		xmlNodePtr class_info = xml_add_node (classes_info, "class");
		xml_set_prop (class_info, "name", i->klass->get_name ().c_str ());
		xml_set_prop (class_info, "level", i->level);
	}
}
	

void
player::save (void) const
{
	xmlDocPtr char_doc = xml_create ("character", "http://www.awemud.net/xml/" BASE_VERSION "/player.dtd");
	if (!char_doc)
		return;

	save (xml_root (char_doc));

	xml_save (char_doc, players::get_path (get_name ().c_str ()));

	xml_free (char_doc);

	message ("Saved player %s\n", name.c_str ());

	return;
}
	
int
player::load (const xmlNodePtr config)
{
	if (character::load (config))
		return 1;

	xmlNodePtr info = xml_find_node (config, "player");
	if (info) {
		xmlNodePtr bday = xml_find_node (info, "birthday");
		if (bday)
			birthday.load (bday);

		if (xml_prop (info, "gm", false))
			SET_FLAG (ply_flags, P_FLAG_GM);
		if (xml_prop (info, "builder", false))
			SET_FLAG (ply_flags, P_FLAG_BUILDER);
		if (xml_prop (info, "admin", false))
			SET_FLAG (ply_flags, P_FLAG_ADMIN);

		prompt = xml_find_text (info, "prompt");
	} else {
		return 1;
	}

	string_t race_name = xml_prop (config, "race");
	if (race_name) {
		race = races.find (race_name);
	} else {
		warning ("Player '%s' has invalid race '%s'", get_name ().c_str (), race_name.c_str ());
		return 1;
	}

	xmlNodePtr exp_info = xml_find_node (config, "exp");
	if (exp_info) {
		for (xmlNodePtr child = xml_find_node (exp_info, "eset"); child != NULL; child = xml_next_node (child, "eset")) {
			int e = get_index_of (exp_names, xml_prop (child, "name"));
			if (e >= 0)
				exp[e] = xml_prop (child, "amount", 0);
		}
	}

	level = 0;
	classes.resize (0);
	xmlNodePtr c_info = xml_find_node (config, "classes");
	if (c_info) {
		for (xmlNodePtr child = xml_find_node (c_info, "class"); child != NULL; child = xml_next_node (child, "class")) {
			const char_class *klass = ::classes.find (xml_prop (child, "name"));
			if (klass) {
				grant_class_levels (klass, xml_prop (child, "level", 1));
			} else {
				warning ("Player '%s' has invalid class '%s'", name.c_str (), xml_prop (child, "name").c_str ());
				return 1;
			}
		}
	} else {
		warning ("Player '%s' has no class info", name.c_str ());
		return 1;
	}

	return 0;
}

int
player::load (const char *name)
{
	/* open the player */
	xmlDocPtr char_doc = xml_open (stringf ("./world/players/%s", name));
	if (char_doc == NULL)
		return 1;

	/* load the player */
	if (load (xml_root (char_doc))) {
		/* damn */
		xml_free (char_doc);
		return 1;
	}
	
	/* the room to enter */
	room *r = NULL;
		
	/* load character-based settings */
	xmlNodePtr ply = xml_find_node (xml_root (char_doc), "player");
	if (ply)
		if (xml_prop (ply, "room"))
			r = server.get_world ()->get_room (xml_prop (ply, "room"));

	/* no room? try the default room */
	if (r == NULL) {
		r = server.get_world ()->get_start_room ();
	}

	message ("Loaded player %s\n", name);

	/* enter the room */
	enter_room (r, NULL);

	/* done */
	xml_free (char_doc);
	return 0;
}

uint
player::get_age (void) const
{
	uint years = server.get_world ()->world_time.year - birthday.year;
	if (server.get_world ()->world_time.month < birthday.month)
		years --;
	else if (server.get_world ()->world_time.month == birthday.month)
		if (server.get_world ()->world_time.day < birthday.day)
			years --;
	return years;
}

void
player::death (void)
{
	do_event ("death", NULL, NULL);
	
	rprint ("%s%s" CNORMAL " has been slain!\n", ncolor (), get_name ().c_str ());

	set_char_flag (C_FLAG_DEAD);
	print ("You have been slain!\n");
	pos = C_LAY;
}

void
player::self_info (void)
{
	print ("Your status:\n");
	print ("  Name:     " CPLAYER "%s" CNORMAL "\n", get_name ().c_str ());
	print ("  Race:     %s\n", race->get_name ().c_str ());
	print ("  Level:    %d\n", level);
	
	/* clsss levels */
	print ("  Class:    " );
	bool did_show = false;
	for (class_list::iterator i = classes.begin (); i != classes.end (); ++i) {
		if (i->level > 0) {
			print ("%s%s (%d)", did_show > 0 ? ", " : "", i->klass->get_name ().c_str (), i->level);
			did_show = true;
		}
	}
	if (!did_show)
		print ("n/a");

	/* exp levels */
	print ("\n  Exp:      " );
	did_show = false;
	for (int i = 0; i < NUM_EXPS; ++ i) {
		if (exp[i] > 0) {
			print ("%s%s (%d)", did_show > 0 ? ", " : "", exp_names[i], exp[i]);
			did_show = true;
		}
	}
	if (!did_show)
		print ("n/a");

	print ("\n  Gender:   %s\n", body->get_gender () == GENDER_MALE ? "male" : body->get_gender () == GENDER_FEMALE ? "female" : "neuter");
	print ("  Age:      %d years (born on %s)\n", get_age (), birthday.date_str ().c_str ());
	print ("  Position: ");
	switch (get_pos ())
	{
		case C_STAND:
			print ("Standing up\n");
			break;
		case C_SIT:
			print ("Sitting down\n");
			break;
		case C_LAY:
			print ("Laying down\n");
			break;
		case C_KNEEL:
			print ("Kneeling\n");
			break;
		default:
			print ("Odd position\n");
			break;
	}
}

void
player::show_inventory (void)
{
	print ("Your inventory:\n");
	if (body)
		body->show (this);
	else
		print (" You have no body!\n");
}

void
player::show_stats (void) {
	print ("Your statistics:\n");
	print ("  Strength:  %s%s" CNORMAL " (%3d)\n", get_stat_color (get_stat(STAT_STRENGTH)), get_stat_level (get_stat (STAT_STRENGTH)), get_stat(STAT_STRENGTH));
	print ("  Agility:   %s%s" CNORMAL " (%3d)\n", get_stat_color (get_stat(STAT_AGILITY)), get_stat_level (get_stat (STAT_AGILITY)), get_stat(STAT_AGILITY));
	print ("  Fortitude: %s%s" CNORMAL " (%3d)\n", get_stat_color (get_stat(STAT_FORTITUDE)), get_stat_level (get_stat (STAT_FORTITUDE)), get_stat(STAT_FORTITUDE));
	print ("  Intellect: %s%s" CNORMAL " (%3d)\n", get_stat_color (get_stat(STAT_INTELLECT)), get_stat_level (get_stat (STAT_INTELLECT)), get_stat(STAT_INTELLECT));
	print ("  Spirit:    %s%s" CNORMAL " (%3d)\n", get_stat_color (get_stat(STAT_SPIRIT)), get_stat_level (get_stat (STAT_SPIRIT)), get_stat(STAT_SPIRIT));
	print ("  Charisma:  %s%s" CNORMAL " (%3d)\n", get_stat_color (get_stat(STAT_CHARISMA)), get_stat_level (get_stat (STAT_CHARISMA)), get_stat(STAT_CHARISMA));
}

void
player::grant_exp (uint type, uint amount) {
	return_if_fail ( type < NUM_EXPS );
	if (amount == 0)
		return;

	uint general = amount / 4;
	amount -= general;

	exp[EXP_GENERAL] += general;
	exp[type] += amount;
}

void
player::grant_class_levels (const char_class *klass, uint levels)
{
	return_if_fail (klass != NULL);

	if (levels == 0)
		return;

	for (class_list::iterator i = classes.begin (); i != classes.end (); ++ i) {
		if (i->klass == klass) {
			i->level += levels;
			level += levels;
			return;
		}
	}

	/* doesn't have level in class */
	class_data cd;
	cd.klass = klass;
	cd.level = levels;
	classes.push_back (cd);
	level += levels;
}

void
player::update (void) {
	if (!is_valid ())
		return;

	character::update ();

	if (round_time == 0 && last_rt > 0) {
		print ("Your round time has ended.\n");
	}
	last_rt = round_time;
}

namespace players
{
	/* helper function to generate path names */
	const char *get_path (const char *name, bool pass)
	{
		static char path[PATH_MAX];
		char *c;

		if (pass)
			snprintf (path, PATH_MAX, "./players/%s.pass", name);
		else
			snprintf (path, PATH_MAX, "./world/players/%s", name);

		for (c = path; *c; ++ c)
			*c = tolower (*c);

		return path;
	}

	bool
	exists (const char *name)
	{
		return does_exist (stringf ("./players/%s.pass", name));
	}

	bool
	check_pass (const char *name, const char *pass)
	{
		char buffer[128];
		const char *path;
		FILE *f;

		path = get_path (name, true);
		f = fopen (path, "r");
		if (f == NULL) {
			return pass ? false : true;
		}

		if (fgets (buffer, 128, f) == NULL) {
			fclose (f);
			return false;
		}

		if (strchr (buffer, '\n'))
			*(strchr (buffer, '\n')) = '\0';

		fclose (f);

		return !strcmp (buffer, crypt (pass, buffer));
	}

	int
	set_pass (const char *name, const char *pass)
	{
		FILE *f;
		char salt[14];
		const char *path;

		/* MD5 Password encryption */
		salt[0] = '$';
		salt[1] = '1';
		salt[2] = '$';
		for (int i = 3; i < 12; i ++) {
			int c = get_random (63);
			if (c < 26) {
				salt[i] = c + 'a';
			} else if (c < 52) {
				salt[i] = c - 26 + 'A';
			} else if (c < 62) {
				salt[i] = c - 52 + '0';
			} else {
				salt[i] = '/';
			}
		}
		salt[12] = '$';
		salt[13] = '\0';

		path = get_path (name, true);
		f = fopen (path, "w");
		if (f == NULL) {
			return errno;
		}

		fprintf (f, "%s\n", crypt (pass, salt));

		fclose (f);

		return 0;
	}

	int
	remove (const char *name)
	{
		const char *path;

		path = get_path (name, false);
		if (does_exist (path))
			unlink (path);

		path = get_path (name, true);
		if (does_exist (path))
			unlink (path);

		return 0;
	}

	/* save a player */
	int
	save (const char *name, const player *ch)
	{
		return_val_if_fail (ch != NULL, 1);
		ch->save ();
		message ("Saved player %s\n", name);
		return 0;
	}

	/* find a player */
	player *
	find (const char *name)
	{
		return_val_if_fail (name != NULL, NULL);
		for (player *ply = ply_list; ply != NULL; ply = ply->p_next ) {
			if (ply->is_valid () && !strcasecmp (name, ply->get_name ()))
				return ply;
		}
		return NULL;
	}

	void
	list_all (player *ply)
	{
		ply->print ("Currently logged in players:\n");
		unsigned long count = 0;
		for (player *conn = ply_list; conn != NULL; conn = conn->p_next) {
			++ count;
			ply->print ("  " CPLAYER "%s", conn->get_name ().c_str ());
			if (conn->is_admin ())
				ply->print (CADMIN " (Admin)" CNORMAL);
			if (conn->is_gm ())
				ply->print (CNORMAL " (GM)");
			if (conn->is_builder ())
				ply->print (CNORMAL " (Builder)");
			ply->print (CNORMAL "\n");
		}
		if (count == 1)
			ply->print ("There is 1 player on-line,\n");
		else
			ply->print ("There are %lu players on-line,\n", count);
	}

	uint
	count (void)
	{
		uint c = 0;
		for (player *ply = ply_list; ply != NULL; ply = ply->p_next)
			if (ply->is_valid ())
				++ c;
		return c;
	}
}

#if 0
int
change_pass::init (void)
{
	return_val_if_fail (conn != NULL, 1);
	return_val_if_fail (conn->get_char () != NULL, 1);

	/* grab out character */
	ch = conn->get_char ();

	conn->toggle_echo (false);

	/* player already has a password? */
	if (players::exists (ch->get_name ())) {
		ch->print ("Changing password...\n");
		set_prompt ("Current password:");
		valid = false;
	} else {
		set_prompt ("New password:");
		valid = true;
	}


	return 0;
}

void
change_pass::close (void)
{
	if (ch->get_conn ())
		ch->get_conn ()->toggle_echo (true);
}

int
change_pass::process (const char *input)
{
	ch->print ("\n");

	/* verified alright?*/
	if (valid) {
		/* set new password already? */
		if (new_pass) {
			/* passwords match? */
			if (input == new_pass) {

				players::set_pass (ch->get_name (), new_pass);
				ch->print ("Password updated.\n");
				message ("Player %s changed password\n", ch->get_name ().c_str ());
				return 1;
			} else {
				ch->print ("Passwords do not match.\n");
				set_prompt ("New password:");
				new_pass = "";
				return 0;
			}
		} else {
			new_pass = input;
			set_prompt ("Verify new password:");
		}
	} else {
		if (players::check_pass (ch->get_name (), input)) {
			valid = true;
			set_prompt ("New password:");
		} else {
			ch->print ("Invalid password.\n");
			return 1;
		}
	}
	return 0;
}
#endif
