/*
  AweMUD - room.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <ctype.h>

#include "room.h"
#include "file.h"
#include "strings.h"
#include "error.h"
#include "mem.h"
#include "color.h"
#include "server.h"
#include "world.h"
#include "player.h"
#include "npc.h"

const char *room_go_acts[] = {
	"go",
	"enter",
	"climb",
	"crawl through",
	NULL
};

const char *room_leave_acts[] = {
	"heads",
	"enters",
	"climbs",
	"crawls through",
	NULL
};

name_registry exit_types;
const_name_id exit_north_id;
const_name_id exit_east_id;
const_name_id exit_south_id;
const_name_id exit_west_id;
const_name_id exit_southeast_id;
const_name_id exit_southwest_id;
const_name_id exit_northeast_id;
const_name_id exit_northwest_id;
const_name_id exit_door_id;
const_name_id exit_portal_id;
const_name_id exit_climb_id;
const_name_id exit_stairs_id;

void
init_room_exits (void) {
	exit_north_id = exit_types ["north"];
	exit_south_id = exit_types ["south"];
	exit_east_id = exit_types ["east"];
	exit_west_id = exit_types ["west"];
	exit_northwest_id = exit_types ["northwest"];
	exit_northeast_id = exit_types ["northeast"];
	exit_southwest_id = exit_types ["southwest"];
	exit_southeast_id = exit_types ["southeast"];
	exit_door_id = exit_types ["door"];
	exit_portal_id = exit_types ["portal"];
	exit_climb_id = exit_types ["climb"];
	exit_stairs_id = exit_types ["stairs"];
}

room *
room_exit::get_room (void) const
{
	return server.get_world()->get_room (target);
}

room_exit *
room_exit::get_exit (void) const
{
	if (!other_exit)
		return NULL;

	room *r = server.get_world()->get_room (target);
	if (r == NULL)
		return NULL;

	room_exit *e = r->get_exit (other_exit);

	return e;
}

bool
room_exit::is_valid (void) const
{
	return name && target && server.get_world()->get_room (target);
}

void
room_exit::save (xmlNodePtr info) const
{
	return_if_fail (info != NULL);

	if (type)
		xml_set_prop (info, "type", exit_types[type]);
	if (is_hidden ())
		xml_set_prop (info, "hidden", 1);
	if (is_door ()) {
		if (is_closed ())
			xml_set_prop (info, "closed", 1);
		if (is_locked ())
			xml_set_prop (info, "locked", 1);
		if (!is_synced () && is_door ())
			xml_set_prop (info, "nosync", 1);
	}

	if (name)
		xml_add_text (info, "name", name);
	if (desc)
		xml_add_text (info, "desc", desc);
	if (target)
		xml_add_text (info, "target", target);
	if (other_exit)
		xml_add_text (info, "otherexit", other_exit);
	if (enter)
		xml_add_text (info, "enter", enter);
	if (leave)
		xml_add_text (info, "leave", leave);
	if (go)
		xml_add_text (info, "go", go);
}

void
room_exit::load (xmlNodePtr info)
{
	return_if_fail (info != NULL);

	string_t tname = xml_prop (info, "type");
	if (tname)
		type = exit_types[tname];
	else
		type = 0;

	flags = 0;
	set_hidden (xml_prop (info, "hidden", false));
	if (type == exit_door_id) {
		set_closed (xml_prop (info, "closed", false));
		set_locked (xml_prop (info, "locked", false));
		set_synced (!xml_prop (info, "nosync", false));
	}

	name = xml_find_text (info, "name");
	if (!name)
		name = exit_types[type];

	desc = xml_find_text (info, "desc");
	target = xml_find_text (info, "target");
	other_exit = xml_find_text (info, "otherexit");
	enter = xml_find_text (info, "enter");
	leave = xml_find_text (info, "leave");
	go = xml_find_text (info, "go");
}

bool
room_exit::handles_type (int gtype) const {
	switch (gtype) {
		case ROOM_EXIT_GO:
			if (type != exit_climb_id)
				return true;
			break;
		case ROOM_EXIT_ENTER:
			if (type == exit_door_id || type == exit_portal_id)
				return true;
			break;
		case ROOM_EXIT_CLIMB:
			if (type == exit_climb_id || type == exit_stairs_id)
				return true;
			break;
	}

	return false;
}

int
room_exit::preferred_type (void) const {
	if (type == exit_climb_id || type == exit_stairs_id) {
		return ROOM_EXIT_CLIMB;
	}
	if (type == exit_door_id || type == exit_portal_id) {
		return ROOM_EXIT_ENTER;
	}

	return ROOM_EXIT_GO;
}

void
room_exit::open (void) {
	flags &= ~EXIT_CLOSED;

	if (is_synced ()) {
		room_exit *other = get_exit ();
		if (other) {
			other->flags &= ~EXIT_CLOSED;
			room *oroom = get_room ();
			if (oroom)
				oroom->print (stringf (CREXIT "%s" CNORMAL " is opened from the other side.\n", other->get_name ().c_str ()));
		}
	}
}

void
room_exit::close (void) {
	flags |= EXIT_CLOSED;

	if (is_synced ()) {
		room_exit *other = get_exit ();
		if (other) {
			other->flags |= EXIT_CLOSED;
			room *oroom = get_room ();
			if (oroom)
				oroom->print (stringf (CREXIT "%s" CNORMAL " is closed from the other side.\n", other->get_name ().c_str ()));
		}
	}
}

void
room_exit::unlock (void) {
	flags &= ~EXIT_LOCKED;

	if (is_synced ()) {
		room_exit *other = get_exit ();
		if (other) {
			other->flags &= ~EXIT_LOCKED;
			room *oroom = get_room ();
			if (oroom)
				oroom->print (stringf ("A click eminates from " CREXIT "%s" CNORMAL ".\n", other->get_name ().c_str ()));
		}
	}
}

void
room_exit::lock (void) {
	flags |= EXIT_LOCKED;

	if (is_synced ()) {
		room_exit *other = get_exit ();
		if (other) {
			other->flags |= EXIT_LOCKED;
			room *oroom = get_room ();
			if (oroom)
				oroom->print (stringf ("A click eminates from " CREXIT "%s" CNORMAL ".\n", other->get_name ().c_str ()));
		}
	}
}

/* constructor */
room::room (void) : area ()
{
	/* clear de values */
	flags = 0;
	outdoors = 0;
	next = NULL;
	exits = NULL;
}

room::~room (void) {
	room_exit *enext;
	
	while (exits != NULL) {
		enext = exits->next;
		delete (exits);
		exits = enext;
	}
}

/* load the freaken room */
int
room::load (const xmlNodePtr info)
{
	return_val_if_fail (info != NULL, 1);

	/* set flags */
	flags = 0;

	title = xml_find_text (info, "title");

	outdoors = xml_prop (info, "outdoors", true);

	if (area::load (info))
		return 1;

	if (!name || !title)
	{
		warning ("Incomplete room (%s,%s)", name.c_str (), title.c_str ());
		return 1;
	}

	short_desc = xml_find_text (info, "shortdesc");

	if (isdigit (name.at (0)))
	{
		warning ("Invalid name: cannot start with a number\n");
		return 1;
	}

	if (server.get_world ()->get_room (name))
	{
		warning ("Duplicate room name: %s", name.c_str ());
		return 1;
	}

	for (xmlNodePtr exit_info = xml_find_node (info, "exit"); exit_info != NULL; exit_info = xml_next_node (exit_info, "exit"))
	{
		room_exit *e = new room_exit ();;
		e->load (exit_info);
		e->next = exits;
		exits = e;
	}

	for (xmlNodePtr object_info = xml_find_node (info, "object"); object_info != NULL; object_info = xml_next_node (object_info, "object"))
	{
		object *obj = new object ();
		obj->load (object_info);
		objects.add (obj);
	}

	for (xmlNodePtr object_info = xml_find_node (info, "o_id"); object_info != NULL; object_info = xml_next_node (object_info, "o_id"))
	{
		string_t s_id = xml_prop (object_info, "id");
		if (s_id) {
			resource_id id;
			id.set_handle (s_id);
			if (id) {
				object *obj = new object ();
				obj->load (id);
				objects.add (obj);
			}
		}
	}

	for (xmlNodePtr character_info = xml_find_node (info, "character"); character_info != NULL; character_info = xml_next_node (character_info, "character"))
	{
		character *ch = new npc ();
		if (!ch->load (character_info)) {
			ch->enter_room (this, NULL);
		} else {
			ch->unref ();
		}
	}

	for (xmlNodePtr character_info = xml_find_node (info, "c_id"); character_info != NULL; character_info = xml_next_node (character_info, "c_id"))
	{
		string_t s_id = xml_prop (character_info, "id");
		if (s_id) {
			resource_id id;
			id.set_handle (s_id);
			if (id) {
				npc *ch = new npc ();
				if (!ch->load (id)) {
					ch->enter_room (this, NULL);
				} else {
					ch->unref ();
				}
			}
		}
	}

	return 0;
}

/* save the stupid thing */
void
room::save (xmlNodePtr info) const
{
	return_if_fail (info != NULL);

	xml_add_text (info, "title", title);
	if (short_desc)
		xml_add_text (info, "shortdesc", short_desc);

	xml_set_prop (info, "outdoors", outdoors);

	area::save (info);

	for (room_exit *e = exits; e != NULL; e = e->next) {
		e->save (xml_add_node (info, "exit"));
	}

	for (const object_list::item *i = objects.entries (); i != NULL; i = i->next) {
		xmlNodePtr o_info = xml_add_node (info, "o_id");
		xml_set_prop (o_info, "id", i->ent->get_id ().get_handle ());
		i->ent->save ();
	}

	for (const character_list::item *i = chars.entries (); i != NULL; i = i->next) {
		if (PLAYER (i->ent)) {
			players::save (i->ent->get_name (), PLAYER(i->ent));
		} else {
			xmlNodePtr c_info = xml_add_node (info, "c_id");
			xml_set_prop (c_info, "id", i->ent->get_id ().get_handle ());
			i->ent->save ();
		}
	}
}

void
room::save (void) const {
	xmlDocPtr r_doc = xml_create ("room", "http://www.awemud.net/xml/" BASE_VERSION "/room.dtd");
	if (!r_doc)
		return;

	save (xml_root (r_doc));

	xml_save (r_doc, stringf ("./world/rooms/%s.room", get_name ().c_str ()));

	xml_free (r_doc);

	return;
}

/* warning - room save based on unique name, not ID */
int
room::load (const char *lname) {
	xmlDocPtr r_doc = xml_open (stringf ("world/rooms/%s.room", lname));
	if (!r_doc)
		return 1;

	int ok = load (xml_root (r_doc));

	xml_free (r_doc);

	return ok;
}

room_exit *
room::get_exit (const char *e_name)
{
	return_val_if_fail (e_name, NULL);
	for (room_exit *i = exits; i != NULL; i = i->next)
		if (i->get_name () == e_name)
			return i;
	return NULL;
}

room_exit *
room::find_exit (const char *e_name)
{
	return_val_if_fail (e_name, NULL);
	for (room_exit *i = exits; i != NULL; i = i->next)
		if (phrase_match (i->get_name ().c_str (), e_name))
			return i;
	return NULL;
}

room_exit *
room::find_exit_type (name_id type)
{
	return_val_if_fail (type, NULL);
	for (room_exit *i = exits; i != NULL; i = i->next)
		if (i->get_type () == type)
			return i;
	return NULL;
}

void
room::announce (string_t text, flags_type flags)
{
	if (CHECK_FLAG (flags, AN_INDOORS_ONLY) && is_outdoors())
		return;
	if (CHECK_FLAG (flags, AN_OUTDOORS_ONLY) && !is_outdoors ())
		return;
	print (text);
}

/* update: one game tick */
void
room::update (void)
{
}

/* print out room */
void
room::show (character *sch, bool shortd)
{
	return_if_fail (sch != NULL);

	sch->print ("[" CRTITLE "%s" CNORMAL "]\n", title.c_str ());
	if (shortd && get_short_desc ().length ()) 
		sch->print ("  %s\n", get_short_desc ().c_str ());
	else
		sch->print ("  %s\n", get_desc ().c_str ());

	if (is_outdoors ()) {
		sch->print ("%s  ", server.get_world ()->get_weather_desc ().c_str ());
		if (server.get_world ()->world_time.is_day ()) {
			sch->print ("The Sun is casting its light about the world.\n");
		} else {
			sch->print ("The moon is making its trek across the night sky.\n");
		}
	}

	if (exits != NULL) {
		bool did_out = false;
		for (room_exit *i = exits; i != NULL; i = i->next)
			if (!i->is_hidden ()) {
				if (!did_out) {
					sch->print ("Visible exits: ");
				}
				sch->print ("%s%s" CREXIT "%s" CNORMAL "%s", did_out ? ", " : "", i->is_closed () ? "[" : "", i->get_name ().c_str (), i->is_closed () ? "]" : "");
				did_out = true;
			}
		if (did_out)
			sch->print (".\n");
	}

	if (chars.entries () != NULL) {
		bool did_out = false;
		for (const character_list::item *i = chars.entries (); i != NULL; i = i->next)
			if (i->ent != sch) {
				if (!did_out) {
					sch->print ("Also in the room: ");
				}
				sch->print ("%s%s%s" CNORMAL, did_out ? ", " : "", i->ent->ncolor (), i->ent->get_name ().c_str ());
				did_out = true;
			}
		if (did_out)
			sch->print (".\n");
	}

	if (objects.entries () != NULL) {
		bool did_out = false;
		for (const object_list::item *i = objects.entries (); i != NULL; i = i->next)
			if (!i->ent->is_hidden ()) {
				if (!did_out) {
					sch->print ("You also see: ");
				}
				sch->print ("%s" CITEM "%s" CNORMAL, did_out ? ", " : "", i->ent->get_name ().c_str ());
				did_out = true;
			}
		if (did_out)
			sch->print (".\n");
	}
}

/* broadcast a message to the room */
void
room::print (const char *msg)
{
	return_if_fail (msg != NULL);
	for (character_list::item *ch = chars.entries (); ch != NULL; ch = ch->next) {
		ch->ent->print ("%s", msg);
	}
}

/* say something */
void
room::say (char const *text, character const *s_ch)
{
	return_if_fail (text != NULL);
	return_if_fail (s_ch != NULL);

	// don't say nothing
	if (!strlen (text))
		return;

	char last_char = text[strlen (text) - 1];
	for (character_list::item *ch = chars.entries (); ch != NULL; ch = ch->next)
		ch->ent->print (CPLAYER "%s" CNORMAL " %s, \"" CTALK "%s" CNORMAL "\"\n", s_ch->get_name ().c_str (), (last_char == '?') ? "asks" : (last_char == '!') ? "exclaims" : "says", text);
}

/* find a character by name */
character *
room::find_char (string_t cname)
{
	return chars.match (cname);
}

/* find an object by name */
object *
room::find_obj (string_t oname)
{
	return objects.match (oname);
}
