/*
  AweMUD - scriptix.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#include <stdlib.h>
#include <dirent.h>
#include <fnmatch.h>

#include "scriptix.h"
#include "server.h"
#include "error.h"
#include "player.h"

SX_SYSTEM *scriptix = NULL;
SX_MODULE *sx_awemud = NULL;
SX_CLASS *sx_entity_class = NULL;

/* callback functions */

void
handle_scriptix_error (const char *str) {
	warning ("Scriptix error: %s", str);
}

int
scriptix_print_hook (const char *, ...) {
	/* Do nothing */
	return 0;
}

/* global functions */

SX_DEFINE_CFUNC(scriptix_log) {
	unsigned int i;
	SX_VALUE *value;
	string_t out;

	for (i = 0; i < sx_argc; ++ i) {
		value = sx_to_str (sx_thread->system, sx_get_value (sx_thread, i));
		if (SX_ISSTRING(sx_thread->system,value) && SX_TOSTRING(value)->len > 0) {
			out += ((char *)SX_TOSTRING(value)->str);
		}
	}

	if (out) {
		message ("%s\n", out.c_str ());
	}
}

SX_DEFINE_CFUNC(scriptix_get_player) {
	player *ply;
	SX_VALUE *name;

	if (sx_argc != 1) {
		sx_raise_error (sx_thread, sx_ArgumentError, NULL);
		return;
	}

	name = sx_get_value (sx_thread, 0);
	if (!SX_ISSTRING (sx_thread->system, name)) {
		sx_raise_error (sx_thread, sx_ArgumentError, NULL);
		return;
	}

	ply = players::find (SX_TOSTRING (name)->str);
	if (ply) {
		sx_push_value (sx_thread, sx_new_entity (ply));
	}
}

/* entity function s*/

void
sx_del_entity (SX_SYSTEM *system, sx_entity *ent) {
	if (ent->ent)
		ent->ent->unref ();
	delete (ent);
}

SX_DEFINE_CFUNC(sx_entity_get_name) {
	sx_push_value (sx_thread, sx_new_str (sx_thread->system, ((sx_entity *)sx_self)->ent->get_name ()));
}

SX_DEFINE_CFUNC(sx_entity_set_name) {
	if (sx_argc != 1) {
		sx_raise_error (sx_thread, sx_ArgumentError, NULL);
		return;
	}

	SX_VALUE *name = sx_get_value (sx_thread, 0);
	if (SX_ISSTRING (sx_thread->system, name)) {
		((sx_entity *)sx_self)->ent->set_name (SX_TOSTRING (name)->str);
	} else {
		sx_raise_error (sx_thread, sx_ArgumentError, NULL);
		return;
	}

	sx_push_value (sx_thread, name); 
}

SX_DEFINE_CFUNC(sx_entity_print) {
	unsigned int i;
	SX_VALUE *value;
	string_t out;

	if (!CHARACTER (((sx_entity *)sx_self)->ent))
		return;

	for (i = 0; i < sx_argc; ++ i) {
		value = sx_to_str (sx_thread->system, sx_get_value (sx_thread, i));
		if (SX_ISSTRING(sx_thread->system,value) && SX_TOSTRING(value)->len > 0) {
			out.append ((char *)SX_TOSTRING(value)->str);
		}
	}

	if (out) {
		CHARACTER (((sx_entity *)sx_self)->ent)->print ("%s\n", out.c_str ());
	}
}

/* other function s*/

int
init_scriptix (void) {
	DIR *dir;
	struct dirent *d;
	char path[PATH_MAX];
	SX_FUNC *func, *lfunc;

	scriptix = sx_create_system ();
	if (scriptix == NULL) {
		return 1;
	}
	sx_init_stdlib (scriptix);

	sx_awemud = sx_new_module (scriptix, sx_name_to_id ("AweMUD"), NULL);
	if (sx_awemud == NULL) {
		sx_free_system (scriptix);
		return 1;
	}

	scriptix->error_hook = handle_scriptix_error;
	scriptix->print_hook = scriptix_print_hook;

	sx_add_func (sx_awemud, sx_new_cfunc (scriptix, sx_name_to_id ("message"), 1, 1, scriptix_log));
	sx_add_func (sx_awemud, sx_new_cfunc (scriptix, sx_name_to_id ("getplayer"), 1, 0, scriptix_get_player));

	sx_entity_class = sx_new_core_class (scriptix, sx_name_to_id ("Entity"), NULL);
	if (sx_entity_class == NULL) {
		return 1;
	}
	sx_add_class (sx_awemud, sx_entity_class);

	sx_entity_class->core->fdel = (sx_class_del)sx_del_entity;

	sx_add_method (scriptix, sx_entity_class, sx_new_cfunc (scriptix, sx_name_to_id ("getname"), 0, 0, sx_entity_get_name));
	sx_add_method (scriptix, sx_entity_class, sx_new_cfunc (scriptix, sx_name_to_id ("print"), 1, 1, sx_entity_print));

	dir = opendir ("./scripts/init.d/");
	lfunc = func = NULL;
	if (dir != NULL) {
		while ((d = readdir (dir)) != NULL) {
			if (!fnmatch ("*.sx", d->d_name, 0)) {
				snprintf (path, PATH_MAX, "./scripts/init.d/%s", d->d_name);
				message ("Loading: %s\n", path);
				sxp_load_file (sx_awemud, path);
				func = sx_get_func (sx_awemud, sx_name_to_id ("init"));
				if (func != lfunc) {
					sx_create_thread (sx_awemud, func, NULL);
					lfunc = func;
					while (sx_runable (scriptix))
						sx_run (scriptix, 1000);
				}
			}
		}
		closedir (dir);
	}

	return 0;
}

void
run_script (const char *name, const char *args, ...) {
	SX_THREAD *thread;
	SX_FUNC *func;
	SX_VALUE *argv, *value;
	va_list va;

	return_if_fail (name != NULL);

	func = sx_get_func (sx_awemud, sx_name_to_id (name));
	if (func) {
		argv = sx_new_array (scriptix, 0, NULL);
		va_start (va, args);
		for (const char *c = args; c != NULL && *c != '\0'; ++ c) {
			value = NULL;
			switch (*c) {
				case 's':
					value = sx_new_str (scriptix, va_arg (va, char *));
					break;
				case 'i':
					value = sx_new_num (va_arg (va, int));
					break;
				case 'e':
					value = sx_new_entity (va_arg (va, entity *));
					break;
			}
			sx_append (scriptix, argv, value);
		};
		va_end (va);

		thread = sx_create_thread (sx_awemud, func, (SX_ARRAY *)argv);
		if (thread) {
			sx_run_until (scriptix, thread->id);
		}
	}
}

void
close_scriptix (void) {
	sx_free_system (scriptix);
}

SX_VALUE *
sx_new_entity (entity *ent) {
	sx_entity *eval;

	eval = (sx_entity *)sx_malloc (scriptix, sizeof (sx_entity));
	if (eval == NULL) {
		return NULL;
	}

	eval->ent = ent;
	if (ent)
		ent->ref ();

	sx_clear_value (scriptix, &eval->header, sx_entity_class);
	return (SX_VALUE *)eval;
}
