/*
  AweMUD - server.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#include "server.h"

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <termios.h>
#include <arpa/telnet.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <fnmatch.h>

#include "room.h"
#include "file.h"
#include "error.h"
#include "object.h"
#include "strings.h"
#include "mem.h"
#include "parse.h"
#include "rand.h"
#include "world.h"
#include "char.h"
#include "body.h"
#include "emotes.h"
#include "help.h"
#include "scriptix.h"
#include "select.h"
#include "npc.h"
#include "player.h"

server_base server;

/* sockets.h extern definition */
select_t sockets_select;

int init (void);
void sigterm_handler (int);
void sighup_handler (int);

volatile bool signaled_shutdown = false;
volatile bool signaled_reload = false;
bool do_chroot = false;
bool serv_loop_flag = true;

int create_accept_sock (int);

void
close_server_on_exit (void)
{
	server.finish ();
}

int
server_base::parse_args (int argc, char **argv)
{
	/* set path */
	char *path = getenv ("AWEMUD_DATA_PATH");
	if (path == NULL)
		path = ".";

	/* verbose */
	int verbose = -1;
	/* daemon flag */
	bool is_daemon = false;
	/* log file */
	char *log_file = NULL;

	/* I'm lazy, deal with it */
#define ARG_TEXT(name,var) } else if (!strncmp (argv[arg], "--" #name "=", 3 + strlen (#name))) { var=argv[arg] + 3 + strlen (#name);
#define ARG_INT(name,var) } else if (!strncmp (argv[arg], "--" #name "=", 3 + strlen (#name))) { var=satol(argv[arg] + 3 + strlen (#name));

	for (int arg = 1; arg < argc; arg ++) {
		if (argv[arg] == NULL) {
			/* DO NOTHING */
		ARG_TEXT(path,path)
		ARG_TEXT(user,user_name)
		ARG_TEXT(group,group_name)
		ARG_INT(port,accept_port)
		ARG_INT(verbose,verbose)
		ARG_TEXT(pid,pid_file)
		ARG_TEXT(log,log_file)
		} else if (!strcmp (argv[arg], "--chroot"))
			do_chroot = true;
		else if (!strcmp (argv[arg], "--daemon") || !strcmp (argv[arg], "-D"))
			is_daemon = true;
	}

	/* set PID file */
	if (!pid_file.length ())
		pid_file = SERV_PID_FILE;

	if (path)
		if (chdir (path))
			error ("chdir(\"%s\") failed", path);

	/* logging */
	if (log_file)
		log = new logfile (log_file);
	if (log == NULL)
		log = new logfile (fdopen (1, "w"), "standard out");
	if (is_daemon)
	{
		close (0);
		close (1);
		close (2);
		int base_fd = open ("/dev/null", O_RDWR); /* stdin */
		dup (base_fd); /* stdout */
		dup (base_fd); /* stderr */
	}
	if (verbose >= 0)
		log->set_verbose (verbose);

	return 0;
}

int
server_base::init (int argc, char **argv)
{
	world_data = NULL;
	log = NULL;
	accept_port = -1;

	if (parse_args (argc, argv))
		return 1;	

	/* set time */
	time (&start_time);

	/* zero out errno */
	errno = 0;
	
	atexit (close_server_on_exit);

	/* info */
	message ("Logging to %s\n", log->get_file ().c_str ());

	/* write PID */
	message ("Writing PID file %s\n", (const char *)pid_file);
	FILE *pid_f = fopen (pid_file, "w");
	if (pid_f != NULL)
		fprintf (pid_f, "%d\n", getpid ());
	else
		error ("Failed to write PID file %s", pid_file.c_str ());
	fclose (pid_f);

	binit = true;

	/* seting up signal handlers */
	struct sigaction sa;
	memset (&sa, 0, sizeof (sa));
	sa.sa_handler = sigterm_handler;
	if (sigaction (SIGTERM, &sa, NULL))
		error ("sigaction () failed (SIGTERM)");
	memset (&sa, 0, sizeof (sa));
	sa.sa_handler = SIG_IGN;
	if (sigaction (SIGPIPE, &sa, NULL))
		error ("sigaction () failed (SIGPIPE)");
	memset (&sa, 0, sizeof (sa));
	sa.sa_handler = sighup_handler;
	if (sigaction (SIGHUP, &sa, NULL))
		error ("sigaction () failed (SIGHUP)");

	/* load port if not alrady specified */
	if (accept_port < 0)
		accept_port = SERV_PORT;

	/* init build-in data */
	message ("Initializing built-ins\n");
	init_random ();
	init_room_exits ();

	/* load character races */
	message ("Loading races\n");
	xmlDocPtr races_doc;
	races_doc = xml_open ("./data/races");
	if (races_doc == NULL) {
		error ("Couldn't open races: %s", strerror (errno));
		return 1;
	}
	if (races.load (xml_root (races_doc), "race")) {
		error ("Failed to load races");
	}
	xml_free(races_doc);
	
	/* load character classes */
	message ("Loading classes\n");
	xmlDocPtr classes_doc;
	classes_doc = xml_open ("./data/classes");
	if (classes_doc == NULL) {
		error ("Couldn't open classes: %s", strerror (errno));
		return 1;
	}
	if (classes.load (xml_root (classes_doc), "class")) {
		error ("Failed to load classes");
	}
	xml_free(classes_doc);

	/* configure templates */
	npc_templates.set_path ("./templates/npcs");
	obj_templates.set_path ("./templates/objects");

	/* load emote database */
	message ("Loading simple emote database\n");
	if (init_emotes ()) {
		error ("init_emotes() failed");
		return 1;
	}

	/* load help database */
	message ("Loading help information\n");
	if (init_help ()) {
		error ("init_help() failed");
		return 1;
	}

	/* load connenct message */
	message ("Loading miscellaneous data\n");
	if (load_data ())
	{
		error ("load_data () failed");
		return 1;
	}

	/* init Scriptix */
	message ("Initializing Scriptix engine\n");
	if (init_scriptix ()) {
		error ("Failed to initialize Scriptix");
		return 1;
	}

	/* load world */
	message ("Loading world\n");
	if (server.get_world ()->load ())
	{
		error ("Couldn't load world");
		return 1;
	}

	/* setup and config */
	message ("Listening on port %d\n", accept_port);
	if ((accept_sock = create_accept_sock (accept_port)) < 0)
		return 1;

	/* read group/user info */
	struct group *grp = NULL;
	if (group_name.length () && !str_is_long (group_name))
		if ((grp = getgrnam (group_name)) == NULL)
			error ("Couldn't find group %s", group_name.c_str ());
	struct passwd *usr = NULL;
	if (user_name.length () && !str_is_long (user_name))
		if ((usr = getpwnam (user_name)) == NULL)
			error ("Couldn't find user %s", user_name.c_str ());

	/* do chroot jail */
	if (do_chroot)
	{
		message ("Chroot'ing the server\n");
		if (chroot ("./"))
			error ("chroot() failed: %s", strerror (errno));
	}

	/* change user/group */
	if (grp)
	{
		message ("Setting group to %s (%d)\n", grp->gr_name, grp->gr_gid);
		if (setregid (grp->gr_gid, grp->gr_gid))
			error ("Couldn't set group");
	}
	else if (str_is_long (group_name))
	{
		message ("Setting group to %s\n", group_name.c_str ());
		if (setregid (satol (group_name), satol (group_name)))
			error ("Couldn't set group");
	}
	if (usr)
	{
		message ("Setting user to %s (%d)\n", usr->pw_name, usr->pw_uid);
		if (setreuid (usr->pw_uid, usr->pw_uid))
			error ("Couldn't set user");
	}
	else if (str_is_long (user_name))
	{
		message ("Setting user to %s\n", user_name.c_str ());
		if (setreuid (satol (user_name), satol (user_name)))
			error ("Couldn't set user");
	}

	return 0;
}

int
server_base::finish (void)
{
	if (!binit)
		return 1;
	binit = 0;

	message ("Cleaning up\n");

	if (world_data)
		world_data->unref ();
	world_data = NULL;

	if (log)
		delete (log);
	log = NULL;

	close_help ();
	close_emotes ();

	close_scriptix ();

	if (pid_file)
		unlink (pid_file);
	return 0;
}

int
server_base::run (void)
{
	/* init data for select () */
	struct timeval tv;
	struct sockaddr_in address;
	size_t addr_len = sizeof (struct sockaddr_in);
	tv.tv_sec = 0;
	tv.tv_usec = 200000;

	gettimeofday (&last_time, NULL);
	game_ticks = 0;

	message ("Server active\n");

	sockets_select.add (accept_sock);

	while (serv_loop_flag)
	{
		sockets_select.do_select (tv);

		if (sockets_select.check (accept_sock))
		{
			int sock;
			if ((sock = accept (accept_sock, (struct sockaddr *) &address, &addr_len)) < 0)
				error ("accept () failed");
			init_socket (sock);
			dprintf (sock, "AweMUD V"VERSION "\r\n");
			dprintf (sock, "Copyright (C) 2000-2002  AwesomePlay Productions, Inc.\r\n");
			dprintf (sock, "Visit http://www.awemud.net for more details.\r\n");

			/* log connection */
			message ("New connection from %s:%d\n", inet_ntoa (address.sin_addr), ntohs (address.sin_port));
			
			/* create a new connection */
			player *conn = new player ();
			if (conn->connect (sock)) {
				delete (conn);
			} else {
				conn->put (parse::ereplace (connect_message (), NULL, get_world ()));
			}
			conn->handle_output ();
		}

		/* process all connections */
		players::process_all ();

		/* update the server info */
		update ();

		/* flush output */
		players::flush_all ();

		/* check for reload */
		if (signaled_reload == true)
		{
			server.save ();
			server.log_reset ();
			signaled_reload = false;
		}

		/* check for signaled_shutdown */
		if (signaled_shutdown == true)
		{
			signaled_shutdown = false;
			message ("Server received a terminating signal\n");
			save ();
			shutdown ();
		}
	}

	message ("Server deactivated\n");

	return 0;
}

void
init_socket (int f)
{
	int  flags;

	flags = fcntl (f, F_GETFL) | O_NONBLOCK;
	if (fcntl (f, F_SETFL, flags) < 0)
	   error ("fcntl () failed");
}

void
server_base::shutdown (void)
{
	message ("Shutting down server\n");
	serv_loop_flag = false;
}

void
server_base::announce (char const *text, flags_type flags) const
{
	server.get_world ()->announce (stringf (CADMIN "SERVER: %s" CNORMAL "\n", text), flags);
}

void
server_base::save (void) const
{
	if (world_data)
		world_data->save ();
}

void
server_base::update (void)
{
	struct timeval cur_time;
	static uint auto_save_time = 0;
	static uint elapsed_time = 0;

	gettimeofday (&cur_time, NULL);
	uint change_time = ((cur_time.tv_sec - last_time.tv_sec) * 1000000 + (cur_time.tv_usec - last_time.tv_usec)) / 1000;
	elapsed_time += change_time;
	auto_save_time += change_time;

	if (auto_save_time >= 900000)	// 900000 milliseconds, or 15 minutes
	{
		auto_save_time = 0;
		save ();
	}

	last_time = cur_time;

	if (elapsed_time >= MSECS_PER_TICK) {
		entity::update_all ();
		elapsed_time %= MSECS_PER_TICK;
	}
}

/* term signal handler */
void
sigterm_handler (int)
{
	message ("Signal SIGTERM received\n");
	signaled_shutdown = true;
}

/* signal HUP handler */
void
sighup_handler (int)
{
	message ("Signal SIGHUP received\n");
	signaled_reload = true;
}

/* deal with logon message */
int
server_base::load_data (void)
{
	FILE *f;
	char line[1024];

	f = fopen ("./data/logon.txt", "r");
	if (f == NULL) error ("Couldn't open logon message file");
	while (fgets (line, 1024, f))
		logon_message_data += line;
	fclose (f);

	f = fopen ("./data/connect.txt", "r");
	if (f == NULL) error ("Couldn't open connect message file");
	while (fgets (line, 1024, f))
		connect_message_data += line;
	fclose (f);

	f = fopen ("./data/quit.txt", "r");
	if (f == NULL) error ("Couldn't open quit message file");
	while (fgets (line, 1024, f))
		quit_message_data += line;
	fclose (f);

	return 0;
}

void
server_base::log_msg (int level, const char *text)
{
	if (log)
		log->print (level, text);
	else
		fprintf (stderr, "%s\n", text);
}

void
server_base::log_reset (void)
{
	if (log)
	{
		message ("Reopening log file\n");
		log->reset ();
	}
}

int
create_accept_sock (int port)
{
	int sock;
	struct sockaddr_in address;
	int i_opt;

	/* create socket */
	if ((sock = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
		error ("socket () failed");
		return -1;
	}

	/* set reuse timeout thingy */
	i_opt = 1;
	setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &i_opt, sizeof (i_opt));

	/* setup and config */
	address.sin_family = AF_INET;
	address.sin_port = htons (port);
	memset (&address.sin_addr, 0, sizeof (address.sin_addr));

	/* bind socket */
	if (bind (sock, (struct sockaddr *)&address, sizeof (address))) {
		error ("bind failed (): Port %d is in use; server may already be running", port);
		return -1;
	}

	/* start listening */
	if (listen (sock, 2)) {
		error ("listen () failed");
		return -1;
	}

	return sock;
}

const xmlNodePtr
server_base::get_obj_template (const char *tag)
{
	return_val_if_fail (tag != NULL, NULL);
	const xmlNodePtr doc = obj_templates.get_template (tag);
	if (doc)
		return xml_find_node (doc, "object");
	else
		return NULL;
}

object *
server_base::get_obj_template_copy (const char *tag)
{
	return_val_if_fail (tag != NULL, NULL);
	const xmlNodePtr info = get_obj_template (tag);
	if (!info)
		return NULL;
	object *obj = new object ();
	if (obj->load (info))
	{
		obj->unref ();
		return NULL;
	}
	return obj;
}

const xmlNodePtr
server_base::get_char_template (const char *tag)
{
	return_val_if_fail (tag != NULL, NULL);
	const xmlNodePtr doc = npc_templates.get_template (tag);
	if (doc)
		return xml_find_node (doc, "character");
	else
		return NULL;
}

character *
server_base::get_char_template_copy (const char *tag)
{
	return_val_if_fail (tag != NULL, NULL);
	const xmlNodePtr info = get_char_template (tag);
	if (!info)
		return NULL;
	character *ch = new npc ();
	if (ch->load (info))
	{
		ch->unref ();
		return NULL;
	}
	return ch;
}
