/*
  AweMUD - strings.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>

#include <vector>

#include "types.h"
#include "strings.h"
#include "error.h"
#include "mem.h"
#include "log.h"

const char *yes_no[3] = {"yes","no",NULL};
const char *on_off[3] = {"on","off",NULL};
const char *a_an[3] = {"a","an",NULL};

char *_s_strtok_buffer = NULL;
char *_s_strtok_cur = NULL;

const char *
strtok_t::strtok (const char *sep)
{
	if (buf)
		free (buf);
	buf = NULL;

	if (!tok) return NULL;

	if (sep)
	{
		for (register char *ptr = tok; *ptr; ptr ++)
		{
			if (*ptr == *sep)
			{
				buf = (char *)malloc (ptr - tok + 1);
				memcpy (buf, tok, ptr - tok);
				buf[ptr - tok] = 0;
				strcpy (tok, ptr + 1);
				return buf;
			}
		}
	}

	buf = tok;
	tok = NULL;

	return buf;
}

char *
suprintf (char const *format, ...)
{
	va_list va;
	char *ret;

	return_val_if_fail (format != NULL, NULL);

	va_start (va, format);
	ret = vsuprintf (format, va);
	va_end (va);

	return ret;
}

char *
vsuprintf (char const *format, va_list va)
{
	va_list va2;
	char *ret;
	int size;
	char buf[2];

	return_val_if_fail (format != NULL, NULL);

	__va_copy (va2, va);
	
	size = vsnprintf (buf, 2, format, va2);
	va_end (va2);

	ret = (char *)malloc (size + 1);
	if (!ret) NO_MEM (ret,size);

	vsnprintf (ret, size + 1, format, va);

	return ret;
}

bool
str_is_long (const char *str)
{
	/* this isn't an error - is valid to be NULL */
	if (str == NULL)
		return false;

	/* no length?  bad */
	if (!*str)
		return false;

	/* a starting + or - ?  good */
	if (*str == '-' || *str == '+')
		str ++;

	/* scan */
	while (*str && isdigit (*str)) str ++;

	/* stopped before end of line?  bad */
	if (*str)
		return false;
		
	/* finished at end of line?  good */
	return true;
}

bool
str_case_equal (const char *str_a, const char *str_b)
{
	if (str_a == str_b)
		return true;
	if (!str_a || !str_b)
		return false;
	if (!strcmp (str_a, str_b))
		return true;
	return false;
}

int
phrase_match (string_t match, string_t test)
{
	if (!match || !test)
		return false;

	struct chunk {
		const char *base;
		int len;
	} chunks[5];
	int cchunk = 0;

	const char *c = test;
	/* skip leading spaces */
	while (*c != '\0' && isspace (*c))
		++ c;
	if (*c == '\0')
		return false;

	/* break up test string */
	cchunk = 1;
	chunks[cchunk - 1].base = c;
	while (*c != '\0') {
		/* end of chunk */
		if (isspace (*c)) {
			/* break at max */
			if (cchunk == 5)
				break;

			/* close current */
			chunks[cchunk - 1].len = c - chunks[cchunk - 1].base;

			/* find next */
			while (*c != '\0' && isspace (*c))
				++ c;
			if (*c == '\0')
				break;
			++ cchunk;
			chunks[cchunk - 1].base = c;
		} else
			++ c;
	}
	/* close last on on stack */
	chunks[cchunk - 1].len = c - chunks[cchunk - 1].base;

	/* compare chunks to match string */
	int matches = 0;
	c = match;
	const char *word = c;
	while (*c != '\0') {
		/* end of chunk */
		if (isspace (*c)) {
			/* compare current chunks to this word */
			if (chunks[matches].len <= c - word && !strncasecmp (word, chunks[matches].base, chunks[matches].len)) {
				/* match */
				++ matches;
				if (matches == cchunk)
					return true;
					
			}

			/* find next word */
			while (*c != '\0' && isspace (*c))
				++ c;
			if (*c == '\0')
				break;
			word = c;
		} else
			++ c;
	}
	if (chunks[matches].len <= c - word && !strncasecmp (word, chunks[matches].base, chunks[matches].len)) {
		/* match */
		++ matches;
	}

	/* matched all chunks? */
	return matches == cchunk ? true : false;
}

long
satol (const char *str)
{
	if (str)
		return strtol (str, NULL, 10);
	return 0;
}

const char *
get_num_suffix (unsigned int num) {
	if (num == 11 || num == 12 || num == 13) { return "th"; }
	num %= 10;
	return num == 1 ? "st" : num == 2 ? "nd" : num == 3 ? "rd" : "th";
}

string_t
timetostr (const time_t &t)
{
	uint day, hour, min, sec;

	sec = t;

	day = sec / (60 * 60 * 24);
	sec = sec % (60 * 60 * 24);

	hour = sec / (60 * 60);
	sec = sec % (60 * 60);

	min = sec / 60;
	sec = sec % 60;

	return stringf ("%u days, %u hours, %u minutes, and %u seconds", day, hour, min, sec);
}

const char *
get_text_of (const char **list, unsigned int index)
{
	return_val_if_fail (list != NULL, NULL);

	for (uint i = 0; i <= index; i ++)
		if (list[i] == NULL)
			return NULL;

	return list[index];
}

int
get_index_of (const char **list, const char *str)
{
	return_val_if_fail (list != NULL, -1);
	return_val_if_fail (str != NULL, -1);

	for (int i = 0; list[i] != NULL; i ++)
		if (!strcasecmp (list[i], str))
			return i;
	return -1;
}

int
guess_index_of (const char **list, const char *str)
{
	return_val_if_fail (list != NULL, -1);
	return_val_if_fail (str != NULL, -1);

	int sl = strlen (str);

	for (int i = 0; list[i] != NULL; i ++)
		if (!strncasecmp (list[i], str, sl))
			return i;
	return -1;
}

int
get_index_of (const char **list, const char *str, int def)
{
	return_val_if_fail (list != NULL, def);

	if (str == NULL)
		return def;

	for (int i = 0; list[i] != NULL; i ++)
		if (!strcasecmp (list[i], str))
			return i;
	return def;
}

int
guess_index_of (const char **list, const char *str, int def)
{
	return_val_if_fail (list != NULL, def);

	if (str == NULL)
		return def;

	int sl = strlen (str);

	for (int i = 0; list[i] != NULL; i ++)
		if (!strncasecmp (list[i], str, sl))
			return i;
	return def;
}

string_t
stringf (char const *format, ...)
{
	char buf[1024];

	if (format == NULL)
		return "";

	va_list va;

	va_start (va, format);
	vsnprintf (buf, 1024, format, va);
	va_end (va);

	return buf;
}
