/*
  AweMUD - world.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/


#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "world.h"
#include "strings.h"
#include "mem.h"
#include "file.h"
#include "server.h"
#include "rand.h"
#include "parse.h"
#include "player.h"

world::world (void) : area ()
{
	rooms = NULL;
}

world::~world (void)
{
	room *r = rooms, *r_next = NULL;
	while (r != NULL) {
		r_next = r->next;
		r->unref ();
		r = r_next;
	}
}

/* load a starting room for players */
room *
world::get_start_room (void) {
	room *r = NULL;

	/* try the starting room */
	if (starting_room)
		r = get_room (starting_room);

	/* no? grab the first room we have then */
	if (r == NULL)
		return rooms;
	else
		return r;
}

/* load the world */
int
world::load (xmlNodePtr world_config)
{
	return_val_if_fail (world_config != NULL, 1);

	/* get info */
	author = xml_find_text (world_config, "author");
	version = xml_find_text (world_config, "version");

	if (!author || !version)
	{
		warning ("Invalid world file - no author and/or version.");
		return 1;
	}

	if (area::load (world_config))
		return 1;

	starting_room = xml_find_text (world_config, "start");

	xmlNodePtr time_info = xml_find_node (world_config, "time");
	if (time_info) {
		time_def.load (time_info);
		world_time.load (time_info);
	} else {
		warning ("Invalid world file - no time info");
		return 1;
	}

	/* load weather */
	xmlNodePtr weather_info = xml_find_node (world_config, "weather");
	if (weather_info) {
		if (weather.load (weather_info))
			warning ("Error loading weather");
	} else {
		warning ("Invalid world file - no weather info");
		return 1;
	}

	/* load rooms */
	xmlNodePtr rooms_info = xml_find_node (world_config, "rooms");
	if (rooms_info != NULL) {
		for (xmlNodePtr r_info = xml_find_node (rooms_info, "room"); r_info != NULL; r_info = xml_next_node (r_info, "room")) {
			room *r = new room ();
			if (!r->load (xml_text (r_info)))
				add_room (r);
			else
				r->unref ();
		}
	}

	return 0;
}

int
world::load (void) {
	/* we don't use the id */
	xmlDocPtr wd = xml_open ("./world/description");
	if (!wd)
		return 1;

	int ok = load (xml_root (wd));

	xml_free (wd);

	return ok;
}

/* save a wolrd */
void
world::save (xmlNodePtr world_config) const
{
	/* save info */
	xml_add_text (world_config, "author", author);
	xml_add_text (world_config, "version", version);

	xml_add_text (world_config, "start", starting_room);

	area::save (world_config);

	xmlNodePtr time_info = xml_add_node (world_config, "time");
	time_def.save (time_info);
	world_time.save (time_info);

	if (weather.save (xml_add_node (world_config, "weather")))
		warning ("Error loading weather");

	/* save rooms */
	xmlNodePtr rooms_info = xml_add_node (world_config, "rooms");
	for (const room *r = rooms; r != NULL; r = r->next) {
		r->save ();
		xml_add_text (rooms_info, "room", r->get_name ());
	}
}

void
world::save (void) const {
	xmlDocPtr wd = xml_create ("world", "http://www.awemud.net/xml/" BASE_VERSION "/wd.dtd");
	if (!wd)
		return;

	save (xml_root (wd));

	/* save world description */
	xml_save (wd, "./world/description");
	xml_free (wd);
}

/* find a room */
room *
world::get_room (const char *r_name)
{
	for (room *r = rooms; r != NULL; r = r->next)
		if (r_name == r->get_name ())
			return r;

	return NULL;
}

/* announce to all rooms in a room */
void
world::announce (const char *str, flags_type flags) {
	for (room *r = rooms; r != NULL; r = r->next)
		r->announce (str, flags);
}

/* update the game world */
void
world::update (void) {
	weather.update (1);
	
	/* update time */
	bool is_day = world_time.is_day ();
	world_time.update (1);
	if (is_day && !world_time.is_day ())
		announce ("The Sun sets on the Western horiron.\n");
	if (!is_day && world_time.is_day ())
		announce ("The Sun begins to rise in the East.\n");
}

void
world::add_room (room *r) {
	return_if_fail (r != NULL);
	r->ref ();
	r->next = rooms;
	rooms = r;
}

void
world::del_room (room *r) {
	return_if_fail (r != NULL);

	if (r == rooms) {
		rooms = r->next;
		r->unref ();
		return;
	}

	room *l = rooms;
	while (l->next != r && l != NULL)
		l = l->next;

	if (l != NULL) {
		l->next = r->next;
		r->unref ();
	}
}

void
world::list_rooms (character *ch) const
{
	return_if_fail (ch != NULL);
	for (room *r = rooms; r != NULL; r = r->next)
		ch->print (" %s\n", r->get_name ().c_str ());
}

void
world::parse_char (int c, char *buf) const {
	switch (c) {
		case 'a':
			snprintf (buf, PARSE_RESULT_SIZE, "%s", author.c_str ());
			break;
		case 'e':
			snprintf (buf, PARSE_RESULT_SIZE, "%s", version.c_str ());
			break;
		/* server stuff - has to be in world */
		case 'v':
			strcpy (buf, VERSION);
			break;
		case 'b':
			strcpy (buf, __DATE__ " " __TIME__);
			break;
		case 'p': {
			uint d = (time (NULL) - server.get_start_time ()); 
			uint seconds = d % 60;
			d /= 60;
			uint minutes = d % 60;
			d /= 60;
			uint hours = d % 24;
			d /= 60;
			uint days = d;
			snprintf (buf, PARSE_RESULT_SIZE, "%u days, %u hours, %u minutes, %u seconds", days, hours, minutes, seconds);
			
			break;
		}
		case 'u':
			snprintf (buf, PARSE_RESULT_SIZE, "%u", players::count ());
			break;
		default:
			entity::parse_char (c, buf);
			break;
	}
}
