/*
  AweMUD - xml.cc
  Copyright (c) 2000-2002, AwesomePlay Productions, Inc.
  All rights reserved.
  
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  
   * Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  
  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
  ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
  DAMAGE.
*/

#include <malloc.h>
#include <libxml/parser.h>

#include "xml.h"
#include "error.h"

xmlDocPtr
xml_open (const char *file)
{
	return_val_if_fail (file != NULL, NULL);

	xmlDocPtr doc;

	doc = xmlParseFile (file);

	return doc;
}

void
xml_free (xmlDocPtr doc)
{
	xmlFreeDoc (doc);
}

xmlNodePtr
xml_root (xmlDocPtr doc)
{
	return xmlDocGetRootElement (doc);
}

xmlNodePtr
xml_find_node (xmlNodePtr par, const char *type)
{
	return_val_if_fail (par != NULL, NULL);
	return_val_if_fail (type != NULL, NULL);

	xmlNodePtr new_node = par->children;

	while (new_node != NULL)
	{
		if (new_node->name && str_case_equal ((const char *)new_node->name, type))
			return new_node;
		new_node = new_node->next;
	}

	return NULL;
}

xmlNodePtr
xml_next_node (xmlNodePtr node, const char *type)
{
	return_val_if_fail (node != NULL, NULL);
	return_val_if_fail (type != NULL, NULL);

	node = node->next;

	while (node != NULL)
	{
		if (node->name && str_case_equal ((const char *)node->name, type))
			return node;
		node = node->next;
	}

	return NULL;
}

string_t
xml_prop (xmlNodePtr node, const char *prop)
{
	return_val_if_fail (node != NULL, "");
	return_val_if_fail (prop != NULL, "");

	char *ret = (char *)xmlGetProp (node, (const xmlChar *)prop);
	string_t ret_s;
	if (ret) {
		ret_s = ret;
		free (ret);
	}
	return ret_s;
}

string_t
xml_prop (xmlNodePtr node, const char *prop, const char *def)
{
	return_val_if_fail (node != NULL, def);
	return_val_if_fail (prop != NULL, def);

	char *ret = (char *)xmlGetProp (node, (const xmlChar *)prop);
	if (ret)
	{
		string_t ret_s = ret;
		free (ret);
		return ret_s;
	}
	else
		return def;
}

int
xml_prop (xmlNodePtr node, const char *prop, int def)
{
	return_val_if_fail (node != NULL, def);
	return_val_if_fail (prop != NULL, def);

	char *ret = (char *)xmlGetProp (node, (const xmlChar *)prop);
	if (ret)
	{
		int ret_i = satol (ret);
		free (ret);
		return ret_i;
	}
	else
		return def;
}

string_t
xml_text (xmlNodePtr node)
{
	return_val_if_fail (node != NULL, "");

	char *str = (char *)xmlNodeListGetString(node->doc, node->children, true);
	string_t ret = "";
	if (str != NULL) {
		ret = str;
		free (str);
	}
	return ret;
}

string_t
xml_find_text (xmlNodePtr node, const char *name)
{
	return_val_if_fail (node != NULL, "");
	return_val_if_fail (name != NULL, "");

	xmlNodePtr t_node = xml_find_node (node, name);
	if (t_node)
		return xml_text (t_node);
	else
		return "";
}

xmlDocPtr
xml_create (const char *name, const char *dtd)
{
	return_val_if_fail (name != NULL, NULL);

	xmlDocPtr doc = xmlNewDoc ((xmlChar *)"1.0");
	if (!doc)
		return NULL;

	doc->children = xmlNewDocNode (doc, NULL, (xmlChar *)name, NULL);
	if (!xml_root (doc))
	{
		xml_free (doc);
		warning ("Failed to create child node for XML document");
		return NULL;
	}

	if (dtd != NULL) 
		xmlCreateIntSubset (doc, (xmlChar *)name, NULL, (xmlChar *)dtd);

	return doc;
}

xmlNodePtr
xml_add_node (xmlNodePtr parent, const char *name)
{
	return_val_if_fail (parent != NULL, NULL);
	return_val_if_fail (name != NULL, NULL);

	return xmlNewChild (parent, NULL, (xmlChar *)name, NULL);
}

void
xml_set_prop (xmlNodePtr node, const char *prop, const char *val)
{
	return_if_fail (node != NULL);
	return_if_fail (prop != NULL);
	return_if_fail (val != NULL);

	xmlSetProp (node, (xmlChar *)prop, (xmlChar *)val);
}

void
xml_set_prop (xmlNodePtr node, const char *prop, int val)
{
	return_if_fail (node != NULL);
	return_if_fail (prop != NULL);

	xmlSetProp (node, (xmlChar *)prop, (xmlChar *)stringf ("%d", val).c_str ());
}

int
xml_save (xmlDocPtr doc, const char *file)
{
	return !xmlSaveFormatFile (file, doc, 1);
}

void
xml_set_text (xmlNodePtr node, const char *text)
{
	return_if_fail (node != NULL);
	return_if_fail (text != NULL);

	xmlNodePtr text_node = xmlStringGetNodeList(node->doc, (const xmlChar *)text);
	xmlAddChild (node, text_node);
}

void
xml_add_text (xmlNodePtr node, const char *name, const char *text)
{
	return_if_fail (node != NULL);
	return_if_fail (name != NULL);
	return_if_fail (text != NULL);

	xmlNodePtr t_node = xml_add_node (node, name);
	xml_set_text (t_node, text);
}
